# SPDX-FileCopyrightText: © 2023 Haystack <https://github.com/deepset-ai/haystack.git>
# SPDX-License-Identifier: Apache License 2.0

import logging
import re
from collections.abc import Callable
from typing import Any, Dict, List, Optional, Tuple, Union

from events import Events
from haystack import Answer
from transformers import TextStreamer

from fastrag.agents.agent_step import AgentStep
from fastrag.agents.memory.conversation_memory import ConversationMemory
from fastrag.agents.tool_handlers import IndexingHandler, QueryHandler
from fastrag.agents.utils import Color, clean_for_prompt, print_text, react_parameter_resolver

logger = logging.getLogger(__name__)


class AgentTokenStreamingHandler:
    DONE_MARKER = "[DONE]"

    def __init__(self, events: Events):
        self.events = events

    def __call__(self, token_received, **kwargs) -> str:
        self.events.on_new_token(token_received, **kwargs)
        return token_received


class HFTokenStreamingHandler(TextStreamer):
    def __init__(self, tokenizer, stream_handler):
        super().__init__(tokenizer=tokenizer, skip_prompt=True)
        self.token_handler = stream_handler

    def on_finalized_text(self, token: str, stream_end: bool = False):
        token_to_send = token + "\n" if stream_end else token
        self.token_handler(token_received=token_to_send, **{})


class Tool:
    """
    Agent uses tools to find the best answer. A tool is a pipeline or a node. When you add a tool to an Agent, the Agent
    can invoke the underlying pipeline or node to answer questions.

    You must provide a name and a description for each tool. The name should be short and should indicate what the tool
    can do. The description should explain what the tool is useful for. The Agent uses the description to decide when
    to use a tool, so the wording you use is important.

    :param name: The name of the tool. The Agent uses this name to refer to the tool in the text the Agent generates.
        The name should be short, ideally one token, and a good description of what the tool can do, for example:
        "Calculator" or "Search". Use only letters (a-z, A-Z), digits (0-9) and underscores (_)."
    :param query_handler: The query_handler to run when the tool queries information.
    :param index_handler: The index_handler to run when the tool requires indexing additional data.
    :param description: A description of what the tool is useful for. The Agent uses this description to decide
        when to use which tool. For example, you can describe a tool for calculations by "useful for when you need to

        answer questions about math".
    """

    def __init__(
        self,
        name: str,
        query_handler: QueryHandler,  # create create generic query class
        index_handler: IndexingHandler = None,  # create create generic index class
        description: str = "",
        output_variable: str = "results",
        logging_color: Color = Color.YELLOW,
    ):
        if re.search(r"\W", name):
            raise ValueError(
                f"Invalid name supplied for tool: '{name}'. Use only letters (a-z, A-Z), digits (0-9) and "
                f"underscores (_)."
            )
        self.name = name

        self.query_handler = query_handler
        self.index_handler = index_handler

        self.description = description
        self.output_variable = output_variable
        self.logging_color = logging_color

    def run(self, tool_input: str, params: Optional[dict] = None) -> str:
        # We can only pass params to pipelines but not to nodes
        return self.query_handler.query(query=tool_input)


class ToolsManager:
    """
    The ToolsManager manages tools for an Agent.
    """

    def __init__(
        self,
        tools: Optional[List[Tool]] = None,
        tool_pattern: str = r"Tool:\s*(\w+)\s*Tool Input:\s*(?:\"([\s\S]*?)\"|((?:.|\n)*))\s*",
    ):
        """
        :param tools: A list of tools to add to the ToolManager. Each tool must have a unique name.
        :param tool_pattern: A regular expression pattern that matches the text that the Agent generates to invoke
            a tool.
        """
        self._tools: Dict[str, Tool] = {tool.name: tool for tool in tools} if tools else {}
        self.tool_pattern = tool_pattern
        self.callback_manager = Events(("on_tool_start", "on_tool_finish", "on_tool_error"))
        self.tools_query_history = {k: dict() for k in self._tools}

    def clear_tool_history(self):
        self.tools_query_history = {k: dict() for k in self._tools}

    @property
    def tools(self):
        return self._tools

    def get_tool_names(self) -> str:
        """
        Returns a string with the names of all registered tools.
        """
        return ", ".join(self.tools.keys())

    def get_tools(self) -> List[Tool]:
        """
        Returns a list of all registered tool instances.
        """
        return list(self.tools.values())

    def get_tool_names_with_descriptions(self) -> str:
        """
        Returns a string with the names and descriptions of all registered tools.
        """
        return "\n".join([f"{tool.name}: {tool.description}" for tool in self.tools.values()])

    def run_tool(self, llm_response: str, params: Optional[Dict[str, Any]] = None) -> str:
        tool_result: str = ""
        additional_params = None
        if self.tools:
            tool_name, tool_input = self.extract_tool_details(llm_response)
            if tool_name and tool_input:
                if tool_name not in self.tools:
                    return "The Tool I have specified is not valid, I will re-phrase the Tool name and try again."

                # check if tool was already used:
                if tool_input in self.tools_query_history[tool_name]:
                    return """I have already used this Tool with this Tool Input, I will now use a DIFFERENT Tool and a different Tool Input."""
                else:
                    self.tools_query_history[tool_name][tool_input] = True

                tool: Tool = self.tools[tool_name]
                try:
                    self.callback_manager.on_tool_start(tool_input, tool=tool)
                    tool_result, additional_params = tool.run(tool_input, params)
                    self.callback_manager.on_tool_finish(
                        tool_result,
                        observation_prefix="Observation: ",
                        llm_prefix="Thought: ",
                        color=tool.logging_color,
                        tool_name=tool.name,
                        tool_input=tool_input,
                    )
                except Exception as e:
                    self.callback_manager.on_tool_error(e, tool=self.tools[tool_name])
                    raise e
        return tool_result, additional_params

    def extract_tool_details(self, llm_response: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Parse the tool name and the tool input from the PromptNode response.
        :param llm_response: The PromptNode response.
        :return: A tuple containing the tool name and the tool input.
        """
        tool_match = re.search(self.tool_pattern, llm_response)
        if tool_match:
            tool_name = tool_match.group(1)
            tool_input = tool_match.group(2) or tool_match.group(3)
            return tool_name.strip('" []\n').strip(), tool_input.strip('" \n')
        return None, None


class Agent:
    """
    An Agent answers queries using the tools you give to it. The tools are pipelines or nodes. The Agent uses a large
    language model (LLM) through the PromptNode you initialize it with. To answer a query, the Agent follows this
    sequence:

    1. It generates a thought based on the query.
    2. It decides which tool to use.
    3. It generates the input for the tool.
    4. Based on the output it gets from the tool, the Agent can either stop if it now knows the answer or repeat the
    process of 1) generate thought, 2) choose tool, 3) generate input.

    Agents are useful for questions containing multiple sub questions that can be answered step-by-step (Multi-hop QA)
    using multiple pipelines and nodes as tools.
    """

    def __init__(
        self,
        generator: Any,
        prompt_template: Optional[Union[str, dict]] = None,
        tools_manager: Optional[ToolsManager] = None,
        memory: Optional[ConversationMemory] = None,
        prompt_parameters_resolver: Optional[Callable] = None,
        max_steps: int = 8,
        final_answer_pattern: str = r"Final Answer\s*:\s*(.*)",
        streaming: bool = True,
    ):
        """
         Creates an Agent instance.

        :param generator: The Generator that the Agent uses to decide which tool to use and what input to provide to
        it in each iteration.
        :param prompt_template: A new PromptTemplate or the name of an existing PromptTemplate for the PromptNode. It's
        used for generating thoughts and choosing tools to answer queries step-by-step. If it's not set, the PromptNode's
        default template is used and if it's not set either, the Agent's default `zero-shot-react` template is used.
        :param tools_manager: A ToolsManager instance that the Agent uses to run tools. Each tool must have a unique name.
        You can also add tools with `add_tool()` before running the Agent.
        :param memory: A Memory instance that the Agent uses to store information between iterations.
        :param prompt_parameters_resolver: A callable that takes query, agent, and agent_step as parameters and returns
        a dictionary of parameters to pass to the prompt_template. The default is a callable that returns a dictionary
        of keys and values needed for the React agent prompt template.
        :param max_steps: The number of times the Agent can run a tool +1 to let it infer it knows the final answer.
            Set it to at least 2, so that the Agent can run one a tool once and then infer it knows the final answer.
            The default is 8.
        :param final_answer_pattern: A regular expression to extract the final answer from the text the Agent generated.
        :param streaming: Whether to use streaming or not. If True, the Agent will stream response tokens from the LLM.
        If False, the Agent will wait for the LLM to finish generating the response and then process it. The default is
        True.
        """
        self.max_steps = max_steps
        self.tm = tools_manager or ToolsManager()
        self.memory = memory
        self.callback_manager = Events(
            (
                "on_agent_start",
                "on_agent_step",
                "on_agent_finish",
                "on_agent_final_answer",
                "on_new_token",
            )
        )
        self.generator = generator
        self.set_agent_streamer()

        self.prompt_template = prompt_template
        self.prompt_parameters_resolver = (
            prompt_parameters_resolver if prompt_parameters_resolver else react_parameter_resolver
        )
        self.final_answer_pattern = final_answer_pattern
        self.add_default_logging_callbacks(streaming=streaming)

    def set_agent_streamer(self):
        tokenizer = self.generator.generation_kwargs["streamer"].tokenizer
        stream_handler = AgentTokenStreamingHandler(self.callback_manager)
        self.generator.generation_kwargs["streamer"] = HFTokenStreamingHandler(
            tokenizer, stream_handler
        )

    def add_default_logging_callbacks(
        self, agent_color: Color = Color.GREEN, streaming: bool = False
    ) -> None:
        def on_tool_finish(
            tool_output: str,
            color: Optional[Color] = None,
            observation_prefix: Optional[str] = None,
            llm_prefix: Optional[str] = None,
            **kwargs: Any,
        ) -> None:
            print_text(observation_prefix)  # type: ignore
            print_text(tool_output, color=color)
            print_text(f"\n{llm_prefix}")

        def on_agent_start(**kwargs: Any) -> None:
            agent_name = kwargs.pop("name", "react")
            print_text(f"\nAgent {agent_name} started with {kwargs}\n")

        def on_agent_final_answer(final_answer: Dict[str, Any], **kwargs: Any) -> None:
            pass

        self.tm.callback_manager.on_tool_finish += on_tool_finish
        self.callback_manager.on_agent_start += on_agent_start
        self.callback_manager.on_agent_final_answer += on_agent_final_answer

        if streaming:
            self.callback_manager.on_new_token += lambda token, **kwargs: print_text(
                token, color=agent_color
            )
        else:
            self.callback_manager.on_agent_step += lambda agent_step: print_text(
                agent_step.generator_node_response, end="\n", color=agent_color
            )

    def add_tool(self, tool: Tool):
        """
        Add a tool to the Agent. This also updates the PromptTemplate for the Agent's PromptNode with the tool name.

        :param tool: The tool to add to the Agent. Any previously added tool with the same name will be overwritten.
        """
        if tool.name in self.tm.tools:
            logger.warning(
                "The agent already has a tool named '%s'. The new tool will overwrite the existing one.",
                tool.name,
            )
        self.tm.tools[tool.name] = tool

    def has_tool(self, tool_name: str) -> bool:
        """
        Check whether the Agent has a tool with the name you provide.

        :param tool_name: The name of the tool for which you want to check whether the Agent has it.
        """
        return tool_name in self.tm.tools

    def run(
        self, query: str, max_steps: Optional[int] = None, params: Optional[dict] = None
    ) -> Dict[str, Union[str, List[Answer]]]:
        """
        Runs the Agent given a query and optional parameters to pass on to the tools used. The result is in the
        same format as a pipeline's result: a dictionary with a key `answers` containing a list of answers.

        :param query: The search query
        :param max_steps: The number of times the Agent can run a tool +1 to infer it knows the final answer.
            If you want to set it, make it at least 2 so that the Agent can run a tool once and then infer it knows the
            final answer.
        :param params: A dictionary of parameters you want to pass to the tools that are pipelines.
                       To pass a parameter to all nodes in those pipelines, use the format: `{"top_k": 10}`.
                       To pass a parameter to targeted nodes in those pipelines, use the format:
                        `{"Retriever": {"top_k": 10}, "Reader": {"top_k": 3}}`.
                        You can only pass parameters to tools that are pipelines, but not nodes.
        """
        self.callback_manager.on_agent_start(name="Agent", query=query, params=params)
        agent_step = self.create_agent_step(max_steps)
        try:
            while not agent_step.is_last():
                agent_step = self._step(query, agent_step, params)
        finally:
            self.callback_manager.on_agent_finish(agent_step)
        final_answer = agent_step.final_answer(query=query)
        self.callback_manager.on_agent_final_answer(final_answer)
        return final_answer

    def _step(self, query: str, current_step: AgentStep, params: Optional[dict] = None):
        # plan next step using the LLM
        generator_node_response = self._plan(query, current_step)

        # from the LLM response, create the next step
        next_step = current_step.create_next_step(generator_node_response)
        self.callback_manager.on_agent_step(next_step)

        # run the tool selected by the LLM
        tool_result = (
            self.tm.run_tool(next_step.generator_node_response, params)
            if not next_step.is_last()
            else None
        )

        if tool_result and len(tool_result) == 2:
            observation, additional_params = tool_result
        else:
            observation = tool_result
            additional_params = None

        # save the input, output and observation to memory (if memory is enabled)
        response_to_save = (
            generator_node_response
            if isinstance(generator_node_response, str)
            else generator_node_response[0]
        )
        memory_data = dict(input=query, output=response_to_save, observation=observation)
        if additional_params:
            memory_data["additional_params"] = additional_params

        self.memory.save(data=memory_data)

        # update the next step with the observation
        next_step.completed(observation)
        return next_step

    def apply_params_to_template(self, template_params):
        roles = template_params.pop("memory")
        if isinstance(self.prompt_template, dict):
            # separate system and user/assistant roles
            all_roles = self.prompt_template["system"] + roles + self.prompt_template["chat"]
            prompt_template = self.generator.pipeline.tokenizer.apply_chat_template(
                all_roles, tokenize=False
            )
            prompt = prompt_template.format(**template_params)
        else:
            template_params["memory"] = self.generator.pipeline.tokenizer.apply_chat_template(
                roles, tokenize=False
            )
            prompt = self.prompt_template.format(**template_params)

        # prompt += "Thought: "
        return prompt

    def _plan(self, query, current_step):
        # first resolve prompt template params
        template_params = self.prompt_parameters_resolver(
            query=query, agent=self, agent_step=current_step
        )
        additional_params = self.memory.get_additional_params()

        # invoke via prompt node

        prompt = self.apply_params_to_template(template_params)
        generator_node_response = self.generator.run(prompt=prompt, **additional_params)
        generator_node_response = [clean_for_prompt(r) for r in generator_node_response["replies"]]

        return generator_node_response

    def create_agent_step(self, max_steps: Optional[int] = None) -> AgentStep:
        """
        Create an AgentStep object. Override this method to customize the AgentStep class used by the Agent.
        """
        return AgentStep(
            max_steps=max_steps or self.max_steps, final_answer_pattern=self.final_answer_pattern
        )



import copy
import logging
from typing import List

import torch
import torch.nn.functional as F
from haystack import Document
from torch import nn
from transformers import BatchEncoding
from transformers.models.t5.modeling_t5 import T5Config, T5ForConditionalGeneration, T5Stack

logger = logging.getLogger(__name__)


def get_padded_tensor(ten_list, value=0):
    max_len = max([x.shape[1] for x in ten_list])
    padded_list = []
    for tensor in ten_list:
        if tensor.shape[1] < max_len:
            tensor = F.pad(
                input=tensor, pad=(0, max_len - tensor.shape[1], 0, 0), mode="constant", value=value
            )
        padded_list.append(tensor)
    return padded_list


def tokenization_result_to_tensor(inputs):
    inputs = get_padded_tensor(inputs, value=0)
    return torch.stack(inputs)


def passages_to_tensors(tokenizer, passage_batch, passage_max_len, pad_to_max_length):
    # if padding to the max length, no padding to max passage length, and vice versa.
    padding = True if not pad_to_max_length else "max_length"

    all_input_ids = []
    all_masks = []

    for input_passages in passage_batch:
        current_passage_batch_result = tokenizer(
            input_passages,
            max_length=passage_max_len,
            truncation=True,
            add_special_tokens=False,
            padding=padding,
            return_tensors="pt",
        )

        all_input_ids.append(current_passage_batch_result["input_ids"])
        all_masks.append(current_passage_batch_result["attention_mask"])

    all_input_ids = tokenization_result_to_tensor(all_input_ids)
    all_masks = tokenization_result_to_tensor(all_masks).bool()

    return all_input_ids, all_masks


class FiDConverter:
    def __init__(self, tokenizer_max_len):
        self.max_len = tokenizer_max_len
        logger.info(f"tokenizer max length is:{self.max_len}")
        self.pre_context = "context:"
        self.pre_title = "title:"
        self.pre_question = "question:"

    def __call__(self, tokenizer, query: str, documents: List[Document], top_k=None):
        model_max_len = tokenizer.model_max_length
        if model_max_len > self.max_len:
            logger.info(
                f"Warning!!! You set {self.max_len} as max len for the tokenizer, which is smaller than model_max_length {model_max_len}."
            )

        question = self.pre_question + " " + query

        first_doc = documents[0]
        if first_doc.meta.get("title") is not None:
            formatted_passages = [
                f"{self.pre_title} {c.meta['title']} {self.pre_context} {c.content}"
                for c in documents
            ]
        else:
            formatted_passages = [f"{self.pre_context} {c.content}" for c in documents]

        formatted_passages_with_question = [[question + " " + t for t in formatted_passages]]
        all_input_ids, all_masks = passages_to_tensors(
            tokenizer, formatted_passages_with_question, self.max_len, False
        )
        return BatchEncoding({"input_ids": all_input_ids, "attention_mask": all_masks})


class FusionInDecoderStack(T5Stack):
    def __init__(self, config, embed_tokens=None):
        # init the block stack using the shared embed_tokens
        super().__init__(config, embed_tokens=embed_tokens)
        self.is_encoder = not self.is_decoder

    def check_for_encoder_input_preprocessing(self, input_ids, attention_mask):
        batch_size = None
        if self.is_encoder:
            input_ids_shape = input_ids.shape
            assert len(input_ids_shape) == 3
            passage_count = input_ids_shape[1]
            batch_size = input_ids_shape[0]

            # flatten batch_size and passage dimensions
            input_ids = input_ids.view(batch_size * passage_count, -1)
            attention_mask = attention_mask.view(batch_size * passage_count, -1)

        return input_ids, attention_mask, batch_size

    def get_last_hidden_state(self, output, return_dict):
        return output.last_hidden_state if return_dict else output[0]

    def output_last_hidden_state(self, output, last_hidden_state, return_dict):
        if return_dict:
            output.last_hidden_state = last_hidden_state
            return output
        else:
            return tuple(last_hidden_state, *output[1:])

    def check_for_encoder_output_preprocessing(self, output, return_dict, batch_size):
        if self.is_encoder:
            last_hidden_state = self.get_last_hidden_state(output, return_dict)
            last_hidden_state = last_hidden_state.view(batch_size, -1, last_hidden_state.shape[-1])
            return self.output_last_hidden_state(output, last_hidden_state, return_dict)
        return output

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        inputs_embeds=None,
        head_mask=None,
        cross_attn_head_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        if self.gradient_checkpointing and self.training:
            use_cache = False

        input_ids, attention_mask, batch_size = self.check_for_encoder_input_preprocessing(
            input_ids, attention_mask
        )

        if self.is_decoder and encoder_hidden_states is not None:
            encoder_attention_mask = encoder_attention_mask.view(*encoder_hidden_states.shape[:-1])

        forward_result = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            inputs_embeds=inputs_embeds,
            head_mask=head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        return self.check_for_encoder_output_preprocessing(forward_result, return_dict, batch_size)


class FiDforConditionalGeneration(T5ForConditionalGeneration):
    _keys_to_ignore_on_load_missing = [
        r"encoder.embed_tokens.weight",
        r"decoder.embed_tokens.weight",
        r"lm_head.weight",
    ]
    _keys_to_ignore_on_load_unexpected = [
        r"decoder.block.0.layer.1.EncDecAttention.relative_attention_bias.weight",
    ]

    def __init__(self, config: T5Config):
        # we utilize the same init function as the original T5ForConditionalGeneration,
        # except for using our FusionInDecoderStack
        super().__init__(config)
        self.model_dim = config.d_model

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False
        self.encoder = FusionInDecoderStack(encoder_config, self.shared)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = FusionInDecoderStack(decoder_config, self.shared)

        self.lm_head = nn.Linear(config.d_model, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None



from typing import Any, Dict, List, Literal, Optional

from haystack.components.generators import HuggingFaceLocalGenerator
from haystack.lazy_imports import LazyImport
from haystack.utils import ComponentDevice, Secret
from transformers import AutoConfig, AutoTokenizer

with LazyImport("Run 'pip install -e .[optimum]'") as onnx_import:
    import onnxruntime as ort
    from optimum.onnxruntime import ORTModelForCausalLM, ORTModelForSeq2SeqLM


class ORTGenerator(HuggingFaceLocalGenerator):
    """
    Generator based on a Hugging Face model loaded with ORT.

    This component provides an interface to generate text using a compressed Hugging Face model that runs locally using ORT.

    Usage example:
    ```python

    generator = ORTGenerator(
        model="microsoft/phi-2",
        task="text-generation",
        generation_kwargs={
            "max_new_tokens": 100,
            "temperature": 0.1,
        }
    )

    print(generator.run("Who is the best American actor?"))
    # {'replies': ['John Cusack']}
    ```
    """

    def __init__(
        self,
        model: str = "microsoft/phi-2",
        task: Optional[Literal["text-generation", "text2text-generation"]] = None,
        device: Optional[ComponentDevice] = None,
        token: Optional[Secret] = Secret.from_env_var("HF_API_TOKEN", strict=False),
        generation_kwargs: Optional[Dict[str, Any]] = None,
        huggingface_pipeline_kwargs: Optional[Dict[str, Any]] = None,
        stop_words: Optional[List[str]] = None,
    ):
        """
        Creates an instance of a OpenVINOGenerator.

        :param model: The name or path of a Hugging Face model for text generation,
        :param task: The task for the Hugging Face pipeline.
            Possible values are "text-generation" and "text2text-generation".
            Generally, decoder-only models like GPT support "text-generation",
            while encoder-decoder models like T5 support "text2text-generation".
            If the task is also specified in the `huggingface_pipeline_kwargs`, this parameter will be ignored.
            If not specified, the component will attempt to infer the task from the model name,
            calling the Hugging Face Hub API.
        :param device: The device on which the model is loaded. If `None`, the default device is automatically
            selected. If a device/device map is specified in `huggingface_pipeline_kwargs`, it overrides this parameter.
        :param token: The token to use as HTTP bearer authorization for remote files.
            If the token is also specified in the `huggingface_pipeline_kwargs`, this parameter will be ignored.
        :param generation_kwargs: A dictionary containing keyword arguments to customize text generation.
            Some examples: `max_length`, `max_new_tokens`, `temperature`, `top_k`, `top_p`,...
            See Hugging Face's documentation for more information:
            - [customize-text-generation](https://huggingface.co/docs/transformers/main/en/generation_strategies#customize-text-generation)
            - [transformers.GenerationConfig](https://huggingface.co/docs/transformers/main/en/main_classes/text_generation#transformers.GenerationConfig)
        :param huggingface_pipeline_kwargs: Dictionary containing keyword arguments used to initialize the
            Hugging Face pipeline for text generation.
            These keyword arguments provide fine-grained control over the Hugging Face pipeline.
            In case of duplication, these kwargs override `model`, `task`, `device`, and `token` init parameters.
            See Hugging Face's [documentation](https://huggingface.co/docs/transformers/en/main_classes/pipelines#transformers.pipeline.task)
            for more information on the available kwargs.
            In this dictionary, you can also include `model_kwargs` to specify the kwargs for model initialization:
            [transformers.PreTrainedModel.from_pretrained](https://huggingface.co/docs/transformers/en/main_classes/model#transformers.PreTrainedModel.from_pretrained)
        :param stop_words: A list of stop words. If any one of the stop words is generated, the generation is stopped.
            If you provide this parameter, you should not specify the `stopping_criteria` in `generation_kwargs`.
            For some chat models, the output includes both the new text and the original prompt.
            In these cases, it's important to make sure your prompt has no stop words.
        """
        onnx_import.check()
        super().__init__(
            model=model,
            task=task,
            device=device,
            token=token,
            generation_kwargs=generation_kwargs,
            huggingface_pipeline_kwargs=huggingface_pipeline_kwargs,
            stop_words=stop_words,
        )
        self.ort_classes = [
            ORTModelForCausalLM,
            ORTModelForSeq2SeqLM,
        ]

        model_instance, tokenizer = self.get_ort_model(model, **self.huggingface_pipeline_kwargs)

        self.huggingface_pipeline_kwargs["model"] = model_instance
        self.huggingface_pipeline_kwargs["tokenizer"] = tokenizer

    def get_ort_model(self, model_name_or_path, **kwargs):
        """We also included some additional parameters for the [ort.SessionOptions](https://onnxruntime.ai/docs/api/c/struct_ort_1_1_session_options.html) for loading the model:

        * graph_optimization_level: **str**
        * intra_op_num_threads: **int**
        * session_config_entries: **dict**

        For example:

        ```python
        generator = ORTGenerator(
            model="my/local/path/quantized",
            task="text-generation",
            generation_kwargs={
                "max_new_tokens": 100,
            }
            huggingface_pipeline_kwargs=dict(
                graph_optimization_level="ORT_ENABLE_ALL",
                intra_op_num_threads=6,
                session_config_entries={
                    "session.intra_op_thread_affinities": "3,4;5,6;7,8;9,10;11,12"
                }
            )
        )
        ```
        """
        model_config = AutoConfig.from_pretrained(model_name_or_path)
        ort_model_class = self.get_ort_class(model_config)

        graph_optimization_level = kwargs.pop("graph_optimization_level", "ORT_ENABLE_ALL")

        session_options = ort.SessionOptions()
        session_options.graph_optimization_level = getattr(
            ort.GraphOptimizationLevel, graph_optimization_level
        )
        session_options.intra_op_num_threads = kwargs.pop("intra_op_num_threads", 6)

        session_config_entries = kwargs.pop("session_config_entries", {})

        for session_config_entry_key, session_config_entry_value in session_config_entries.items():
            session_options.add_session_config_entry(
                session_config_entry_key, session_config_entry_value
            )

        self.huggingface_pipeline_kwargs = kwargs

        model = ort_model_class.from_pretrained(
            model_name_or_path, session_options=session_options, **kwargs
        )
        return model, AutoTokenizer.from_pretrained(model_name_or_path)

    def get_ort_class(self, config):
        for current_class in self.ort_classes:
            class_suffix = current_class.__name__.replace("ORTModel", "")

            # If the architecture matches the ORT class, return it
            if config.architectures[0].endswith(class_suffix):
                # return the class and the task
                return current_class

        # No match has been found
        raise Exception("No ORT class found for the model.")



import inspect
import logging
from typing import Any, Dict, List, Optional

import pandas as pd
from haystack import Document, default_from_dict, default_to_dict
from haystack.lazy_imports import LazyImport

with LazyImport("Run 'pip install .[colbert]' to install the ColBERT library.") as colbert_import:
    from colbert import Indexer, Searcher
    from colbert.infra import ColBERTConfig, Run, RunConfig


logger = logging.getLogger(__name__)


class PLAIDDocumentStore:
    """
    Store for ColBERT v2 with PLAID indexing.

    Parameters:

    index_path: directory containing PLAID index files.
    checkpoint_path: directory containing ColBERT checkpoint model files.
    collection_path: a csv/tsv data file of the form (id,content), no header line.

    create: whether to create a new index or load an index from disk. Default: False.

    nbits: number of bits to quantize the residual vectors. Default: 2.
    kmeans_niters: number of kmeans clustering iterations. Default: 1.
    gpus: number of GPUs to use for indexing. Default: 0.
    rank: number of ranks to use for indexing. Default: 1.
    doc_maxlen: max document length. Default: 120.
    query_maxlen: max query length. Default: 60.

    """

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this store to a dictionary.
        """
        params = inspect.signature(self.__init__).parameters
        init_params = {k: getattr(self, k) for k in params}
        return default_to_dict(
            self,
            **init_params,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "PLAIDDocumentStore":
        """
        Deserializes the store from a dictionary.
        """
        return default_from_dict(cls, data)

    def count_documents(self) -> int:
        """
        Returns the number of documents stored.
        """
        return len(self.docs)

    def filter_documents(self, filters: Optional[Dict[str, Any]] = None) -> List[Document]:
        """
        Returns the documents that match the filters provided.
        """
        pass

    def write_documents(self, documents: List[Document], **kwargs) -> int:
        """
        Writes (or overwrites) documents into the DocumentStore, return the number of documents that was written.
        """
        raise Exception(
            "PLAIDDocumentStore can only be used as a read-only store. A new index is needed for adding/changing documents"
        )

    def delete_documents(self, document_ids: List[str]) -> None:
        """
        Deletes all documents with a matching document_ids from the DocumentStore.
        """
        pass

    def __init__(
        self,
        index_path,
        checkpoint_path,
        collection_path,
        create=False,
        nbits=2,
        gpus=0,
        ranks=1,
        doc_maxlen=120,
        query_maxlen=60,
        kmeans_niters=4,
    ):
        colbert_import.check()
        self.index_path = index_path
        self.checkpoint_path = checkpoint_path
        self.collection_path = collection_path
        self.nbits = nbits
        self.gpus = gpus
        self.ranks = ranks
        self.doc_maxlen = doc_maxlen
        self.query_maxlen = query_maxlen
        self.kmeans_niters = kmeans_niters

        if create:
            self._create_index()

        self.docs = pd.read_csv(
            collection_path, sep="\t" if collection_path.endswith(".tsv") else ",", header=None
        )
        self.titles = len(self.docs.columns) > 2
        self._load_index()

    def _load_index(self):
        """Load PLAID index from the paths given to the class and initialize a Searcher object."""
        with Run().context(
            RunConfig(index_root=self.index_path, nranks=self.ranks, gpus=self.gpus)
        ):
            self.store = Searcher(
                index="", collection=self.collection_path, checkpoint=self.checkpoint_path
            )

        logger.info("Loaded PLAIDDocumentStore index")

    def _create_index(self):
        """Generate a PLAID index from a given ColBERT checkpoint.

        Given a checkpoint and a collection of documents, an Indexer object will be created.
        The index will then be generated, written to disk at `index_path` and finally it
        will be loaded.
        """

        with Run().context(
            RunConfig(index_root=self.index_path, nranks=self.ranks, gpus=self.gpus)
        ):
            config = ColBERTConfig(
                doc_maxlen=self.doc_maxlen,
                query_maxlen=self.query_maxlen,
                nbits=self.nbits,
                kmeans_niters=self.kmeans_niters,
            )
            indexer = Indexer(checkpoint=self.checkpoint_path, config=config)
            indexer.index("", collection=self.collection_path, overwrite=True)

        logger.info("Created PLAIDDocumentStore Index.")

    def query(self, query_str, top_k=10) -> List[Document]:
        """
        Query the Colbert v2 + Plaid store.

        Returns: list of Haystack documents.
        """

        doc_ids, _, scores = self.store.search(text=query_str, k=top_k)

        documents = [
            Document.from_dict(
                {
                    "content": self.docs.iloc[_id][1],
                    "id": _id,
                    "score": score,
                    "meta": {"title": self.docs.iloc[_id][2] if self.titles else None},
                }
            )
            for _id, score in zip(doc_ids, scores)
        ]

        return documents



import base64
import logging
from io import BytesIO
from typing import List

import requests
from haystack import component
from haystack.components.builders import PromptBuilder


class MultiModalPromptBuilder(PromptBuilder):
    """
    PromptBuilder is a component that renders a prompt from a template string using Jinja2 templates.
    The template variables found in the template string are used as input types for the component and are all required.

    Usage example:
    ```python
    template = "Translate the following context to {{ target_language }}. Context: {{ snippet }}; Translation:"
    builder = PromptBuilder(template=template)
    builder.run(target_language="spanish", snippet="I can't speak spanish.")
    ```
    """

    @component.output_types(prompt=str, images=List[str])
    def run(self, **kwargs):
        """
        :param kwargs:
            The variables that will be used to render the prompt template.

        :returns: A dictionary with the following keys:
            - `prompt`: The updated prompt text after rendering the prompt template.
        """
        prompt_dict = {"prompt": self.template.render(kwargs)}
        prompt_dict["images"] = [
            self.get_base64_from_url(doc.meta["image_url"]) for doc in kwargs["documents"]
        ]

        return prompt_dict

    def get_base64_from_url(self, image_url):
        response = requests.get(
            image_url,
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
            },
        )
        buffered = BytesIO(response.content)
        img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")
        return img_str


component._component(MultiModalPromptBuilder)



import logging
from typing import Any, Dict, List, Optional, Union

import torch
from haystack import Document, component
from haystack.components.generators import HuggingFaceLocalGenerator
from haystack.utils import ComponentDevice, Secret
from haystack.utils.hf import StopWordsCriteria
from transformers import AutoTokenizer, GenerationConfig, Pipeline, StoppingCriteriaList, pipeline

from .fid_utils import FiDConverter, FiDforConditionalGeneration

logger = logging.getLogger(__name__)


@component
class FiDGenerator(HuggingFaceLocalGenerator):
    def __init__(
        self,
        model: str = "Intel/fid_flan_t5_base_nq",
        task: str = "text2text-generation",
        device: Optional[ComponentDevice] = None,
        token: Optional[Secret] = Secret.from_env_var("HF_API_TOKEN", strict=False),
        generation_kwargs: Optional[Dict[str, Any]] = None,
        huggingface_pipeline_kwargs: Optional[Dict[str, Any]] = {},
        stop_words: Optional[List[str]] = None,
    ):
        super(FiDGenerator, self).__init__(
            model=model,
            task=task,
            device=device,
            token=token,
            generation_kwargs=generation_kwargs,
            huggingface_pipeline_kwargs=huggingface_pipeline_kwargs.copy(),
            stop_words=stop_words,
        )

        tokenizer = AutoTokenizer.from_pretrained(model)
        self.orig_kwargs = huggingface_pipeline_kwargs
        self.huggingface_pipeline_kwargs["tokenizer"] = tokenizer
        self.input_converter = FiDConverter(tokenizer.model_max_length)

    def warm_up(self):
        """
        Initializes the component.
        """
        if self.pipeline is None:
            model_name = self.huggingface_pipeline_kwargs.pop("model")
            model_instance = FiDforConditionalGeneration.from_pretrained(
                model_name, **self.orig_kwargs
            )
            self.huggingface_pipeline_kwargs["model"] = model_instance
            super(FiDGenerator, self).warm_up()

    @component.output_types(replies=List[str])
    def run(
        self,
        prompt: str,
        documents: List[Document],
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ):
        if self.pipeline is None:
            raise RuntimeError(
                "The generation model has not been loaded. Please call warm_up() before running."
            )

        if not prompt:
            return {"replies": []}

        # merge generation kwargs from init method with those from run method
        updated_generation_kwargs = {**self.generation_kwargs, **(generation_kwargs or {})}

        # Input conversion using FiDConverter
        inputs = self.input_converter(self.pipeline.tokenizer, prompt, documents).to(
            self.pipeline.model.device
        )

        gen_outputs = self.pipeline.model.generate(
            **inputs,
            stopping_criteria=self.stopping_criteria_list,
            **updated_generation_kwargs,
        )

        generated_tokens = gen_outputs[0]
        replies = self.pipeline.tokenizer.decode(generated_tokens, skip_special_tokens=True)

        if self.stop_words:
            # the output of the pipeline includes the stop word
            replies = [
                reply.replace(stop_word, "").rstrip()
                for reply in replies
                for stop_word in self.stop_words
            ]

        return {"replies": replies}



from typing import Any, Dict, List, Optional

from haystack import component
from haystack.components.generators import HuggingFaceLocalGenerator
from haystack.lazy_imports import LazyImport
from haystack.utils import ComponentDevice, Secret

with LazyImport("Install ipex using 'pip install -e .[intel]'") as ipex_import:
    import intel_extension_for_pytorch as ipex
    import torch
    from haystack.utils.hf import StopWordsCriteria  # pylint: disable=ungrouped-imports
    from transformers import AutoModelForCausalLM, AutoTokenizer, StoppingCriteriaList


class IPEXGenerator(HuggingFaceLocalGenerator):
    """
    A generator class that uses IPEX (Intel PyTorch Extension) for faster inference on Xeon CPUs.

    Args:
        model_name_or_path (str): The name or path of the pre-trained model. Defaults to "meta-llama/Llama-2-7b-hf".
        device (Optional[ComponentDevice]): The device to run the generator on. Defaults to None.
        token (Optional[Secret]): The secret token for authentication. Defaults to Secret.from_env_var("HF_API_TOKEN", strict=False).
        generation_kwargs (Optional[Dict[str, Any]]): Additional generation arguments. Defaults to None.
        huggingface_pipeline_kwargs (Optional[Dict[str, Any]]): Additional arguments for the Hugging Face pipeline. Defaults to None.
        stop_words (Optional[List[str]]): A list of stop words to remove from the generated replies. Defaults to None.
        dtype (Optional[str]): The data type to use for inference. Defaults to "bfloat16".
        **kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        model_name_or_path: str = "meta-llama/Llama-2-7b-hf",
        device: Optional[ComponentDevice] = None,
        token: Optional[Secret] = Secret.from_env_var("HF_API_TOKEN", strict=False),
        generation_kwargs: Optional[Dict[str, Any]] = None,
        huggingface_pipeline_kwargs: Optional[Dict[str, Any]] = None,
        stop_words: Optional[List[str]] = None,
        dtype: Optional[str] = "bfloat16",  ## bfloat16 of float32
        **kwargs,
    ):
        ipex_import.check()
        super().__init__(
            model=model_name_or_path,
            device=device,
            token=token,
            generation_kwargs=generation_kwargs,
            huggingface_pipeline_kwargs=huggingface_pipeline_kwargs,
            stop_words=stop_words,
        )
        self.model_name_or_path = model_name_or_path
        self.model = None
        self.tokenizer = token
        self.stop_words = stop_words
        self.generation_kwargs = generation_kwargs or {}
        self.amp_enabled = True if dtype != "float32" else False
        self.amp_dtype = getattr(torch, dtype)

    def warm_up(self):
        """
        Initializes the generator by loading the pre-trained model and tokenizer.
        """
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name_or_path,
            torch_dtype=self.amp_dtype,
            low_cpu_mem_usage=True,
            token=self.token,
        ).eval()
        self.model = self.model.to(memory_format=torch.channels_last)
        self.model = ipex.llm.optimize(
            self.model,
            dtype=self.amp_dtype,
            inplace=True,
            deployment_mode=True,
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name_or_path, token=self.token)
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.tokenizer.padding_side = "left"
        self.huggingface_pipeline_kwargs["tokenizer"] = self.tokenizer
        if self.stop_words and self.stopping_criteria_list is None:
            stop_words_criteria = StopWordsCriteria(
                tokenizer=self.pipeline.tokenizer,
                stop_words=self.stop_words,
                device=self.pipeline.device,
            )
            self.stopping_criteria_list = StoppingCriteriaList([stop_words_criteria])

    @component.output_types(replies=List[str])
    def run(
        self,
        prompt: str,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, List[str]]:
        """
        Generates replies based on the given prompt.

        Args:
            prompt (str): The prompt for generating replies.
            generation_kwargs (Optional[Dict[str, Any]]): Additional generation arguments. Defaults to None.

        Returns:
            Dict[str, List[str]]: A dictionary containing the generated replies.
        """
        if self.model is None or self.tokenizer is None:
            raise RuntimeError("Generator is not initialized. Run warm_up() first.")
        if not prompt:
            return {"replies": []}

        updated_generation_kwargs = {**self.generation_kwargs, **(generation_kwargs or {})}

        with torch.inference_mode(), torch.cpu.amp.autocast(enabled=self.amp_enabled):
            inputs = self.tokenizer(prompt, truncation=True, return_tensors="pt")
            gen_ids = self.model.generate(
                **inputs, stopping_criteria=self.stopping_criteria_list, **updated_generation_kwargs
            )
            generated_tokens = gen_ids[0][inputs["input_ids"].shape[1] :]
            replies = self.tokenizer.decode(generated_tokens, skip_special_tokens=True)
        if self.stop_words:
            for stop_word in self.stop_words:
                replies = replies.replace(stop_word, "").rstrip()
        return {"replies": [replies]}



from typing import Optional, Union

from haystack.components.embedders import (
    SentenceTransformersDocumentEmbedder,
    SentenceTransformersTextEmbedder,
)
from haystack.components.embedders.backends.sentence_transformers_backend import (
    _SentenceTransformersEmbeddingBackend,
)
from haystack.lazy_imports import LazyImport
from haystack.utils import Secret

with LazyImport(
    "Run pip install .[intel]' to install Intel optimized frameworks"
) as optimized_intel_import:
    from optimum.intel import IPEXModel


class _IPEXSentenceTransformersEmbeddingBackend(_SentenceTransformersEmbeddingBackend):
    """
    Class to manage Sentence Transformers embeddings.
    """

    def __init__(
        self,
        model: str,
        device: Optional[str] = None,
        auth_token: Optional[Secret] = None,
        trust_remote_code: bool = False,
    ):
        import sentence_transformers

        class _IPEXSTTransformers(sentence_transformers.models.Transformer):
            def _load_model(self, model_name_or_path, config, cache_dir, **model_args):
                print("Loading IPEX ST Transformer model")
                optimized_intel_import.check()
                self.auto_model = IPEXModel.from_pretrained(
                    model_name_or_path, config=config, cache_dir=cache_dir, **model_args
                )
                self.auto_model.eval()

        class _IPEXSentenceTransformer(sentence_transformers.SentenceTransformer):
            def _load_auto_model(
                self,
                model_name_or_path: str,
                token: Optional[Union[bool, str]],
                cache_folder: Optional[str],
                revision: Optional[str] = None,
                trust_remote_code: bool = False,
            ):
                """
                Creates a simple Transformer + Mean Pooling model and returns the modules
                """
                transformer_model = _IPEXSTTransformers(
                    model_name_or_path,
                    cache_dir=cache_folder,
                    model_args={
                        "token": token,
                        "trust_remote_code": trust_remote_code,
                        "revision": revision,
                    },
                    tokenizer_args={
                        "token": token,
                        "trust_remote_code": trust_remote_code,
                        "revision": revision,
                    },
                )
                pooling_model = sentence_transformers.models.Pooling(
                    transformer_model.get_word_embedding_dimension(), "mean"
                )
                return [transformer_model, pooling_model]

            @property
            def device(self):
                return "cpu"

        self.model = _IPEXSentenceTransformer(
            model_name_or_path=model,
            device=device,
            use_auth_token=auth_token.resolve_value() if auth_token else None,
            trust_remote_code=trust_remote_code,
        )


def ipex_model_warm_up(self):
    """
    Initializes the component.
    """
    if not hasattr(self, "embedding_backend"):
        self.embedding_backend = _IPEXSentenceTransformersEmbeddingBackend(
            model=self.model,
            device=self.device.to_torch_str(),
            auth_token=self.token,
        )


class IPEXSentenceTransformersDocumentEmbedder(SentenceTransformersDocumentEmbedder):
    """
    A document embedder that uses IPEX for efficient computation.

    This class extends the base `SentenceTransformersDocumentEmbedder` class and provides an implementation
    that utilizes IPEX for faster document embedding computation.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class IPEXSentenceTransformersTextEmbedder(SentenceTransformersTextEmbedder):
    """
    A text embedder that uses IPEX for efficient text embedding.

    This class extends the `SentenceTransformersTextEmbedder` class and provides
    an implementation that utilizes IPEX for faster and more efficient text embedding.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


IPEXSentenceTransformersDocumentEmbedder.warm_up = ipex_model_warm_up
IPEXSentenceTransformersTextEmbedder.warm_up = ipex_model_warm_up



# SPDX-FileCopyrightText: © 2023 Haystack <https://github.com/deepset-ai/haystack.git>
# SPDX-License-Identifier: Apache License 2.0

import logging
import re
from typing import Any, Dict, Optional


class Answer:
    def __init__(self, answer, type):
        self.answer = answer
        self.type = type


logger = logging.getLogger(__name__)


class AgentStep:
    """
    The AgentStep class represents a single step in the execution of an agent.
    """

    def __init__(
        self,
        current_step: int = 1,
        max_steps: int = 10,
        final_answer_pattern: Optional[str] = None,
        generator_node_response: str = "",
        transcript: str = "",
    ):
        """
        :param current_step: The current step in the execution of the agent.
        :param max_steps: The maximum number of steps the agent can execute.
        :param final_answer_pattern: The regex pattern to extract the final answer from the PromptNode response. If no
        pattern is provided, entire prompt node response is considered the final answer.
        :param generator_node_response: The PromptNode response received.
        text it generated during execution up to this step. The transcript is used to generate the next prompt.
        """
        self.current_step = current_step
        self.max_steps = max_steps
        self.final_answer_pattern = final_answer_pattern or r"^([\s\S]+)$"
        self.generator_node_response = generator_node_response
        self.transcript = transcript

    def create_next_step(self, generator_node_response: Any, current_step: Optional[int] = None):
        """
        Creates the next agent step based on the current step and the PromptNode response.
        :param generator_node_response: The PromptNode response received.
        :param current_step: The current step in the execution of the agent.
        """
        if not isinstance(generator_node_response, list) or not generator_node_response:
            raise Exception(
                f"Agent output must be a non-empty list of str, but {generator_node_response} received. "
                f"Transcript:\n{self.transcript}"
            )
        cls = type(self)
        return cls(
            current_step=current_step if current_step else self.current_step + 1,
            max_steps=self.max_steps,
            final_answer_pattern=self.final_answer_pattern,
            generator_node_response=generator_node_response[0],
            transcript=self.transcript,
        )

    def final_answer(self, query: str) -> Dict[str, Any]:
        """
        Formats an answer as a dict containing `query` and `answers` similar to the output of a Pipeline.
        The full transcript based on the Agent's initial prompt template and the text it generated during execution.

        :param query: The search query
        """
        answer: Dict[str, Any] = {
            "query": query,
            "answers": [Answer(answer="", type="generative")],
            "transcript": self.transcript,
        }
        if self.current_step > self.max_steps:
            logger.warning(
                "Maximum number of iterations (%s) reached for query (%s). Increase max_steps "
                "or no answer can be provided for this query.",
                self.max_steps,
                query,
            )
        else:
            final_answer = self.parse_final_answer()
            if not final_answer:
                logger.warning(
                    "Final answer parser (%s) could not parse response (%s).",
                    self.final_answer_pattern,
                    self.generator_node_response,
                )
            else:
                answer = {
                    "query": query,
                    "answers": [Answer(answer=final_answer, type="generative")],
                    "transcript": self.transcript,
                }
        return answer

    def is_last(self) -> bool:
        """
        Check if this is the last step of the Agent.
        :return: True if this is the last step of the Agent, False otherwise.
        """
        return bool(self.parse_final_answer()) or self.current_step > self.max_steps

    def completed(self, observation: Optional[str]) -> None:
        """
        Update the transcript with the observation
        :param observation: received observation from the Agent environment.
        """
        self.transcript += (
            f"{self.generator_node_response}\nObservation: {observation}\nThought:"
            if observation
            else self.generator_node_response
        )

    def __repr__(self) -> str:
        """
        Return a string representation of the AgentStep object.

        :return: A string that represents the AgentStep object.
        """
        return (
            f"AgentStep(current_step={self.current_step}, max_steps={self.max_steps}, "
            f"generator_node_response={self.generator_node_response}, final_answer_pattern={self.final_answer_pattern}, "
            f"transcript={self.transcript})"
        )

    def parse_final_answer(self) -> Optional[str]:
        """
        Parse the final answer from the response of the prompt node.

        This function searches the prompt node's response for a match with the
        pre-defined final answer pattern. If a match is found, it's returned as the
        final answer after removing leading/trailing quotes and whitespaces.
        If no match is found, it returns None.

        :return: The final answer as a string if a match is found, otherwise None.
        """
        # Search for a match with the final answer pattern in the prompt node response
        final_answer_match = re.search(self.final_answer_pattern, self.generator_node_response)

        if final_answer_match:
            # If a match is found, get the first group (i.e., the content inside the parentheses of the regex pattern)
            final_answer = final_answer_match.group(1)

            # Remove leading/trailing quotes and whitespaces, then return the final answer
            return final_answer.strip('" ')  # type: ignore
        elif "Tool" not in self.generator_node_response:
            # if no tool has been used, return the response
            return self.generator_node_response
        else:
            # If no match is found, return None
            return None



from typing import Optional

from haystack import Document
from tqdm import tqdm

from fastrag.agents.base import Tool


class DocWithImageTool(Tool):
    def run(self, tool_input: str, params: Optional[dict] = None) -> str:
        result = self.query_handler.query(tool_input, params)

        result_dict = result["prompt_builder"]
        text = result_dict["prompt"]

        tool_result = f"""
Success! I have found the image and text I need:
{text}
I will now use the image and text to answer the question.
"""
        del result_dict["prompt"]
        additional_params = result_dict
        return tool_result, additional_params

    def upload_data_to_pipeline(self, params):
        """
        Upload documents into the textual with images document store.
        """
        # ensure that params contain the text to upload
        if "data_rows" not in params:
            return

        for data_rows in params["data_rows"]:
            docs = [
                Document(
                    content=ex["content"], meta={"title": ex["title"], "image_url": ex["image_url"]}
                )
                for ex in tqdm(data_rows)
            ]
            self.index_handler.index(docs)


class DocTool(Tool):
    def run(self, tool_input: str, params: Optional[dict] = None) -> str:
        result = self.query_handler.query(tool_input, params)

        result_dict = result["prompt_builder"]
        text = result_dict["prompt"]

        tool_result = text

        return tool_result

    def upload_data_to_pipeline(self, params):
        """
        Upload documents into the textual with images document store.
        """
        # ensure that params contain the text to upload
        if "data_rows" not in params:
            return

        for data_rows in params["data_rows"]:
            docs = [
                Document(content=ex["content"], meta={"title": ex["title"]})
                for ex in tqdm(data_rows)
            ]
            self.index_handler.index(docs)


TOOLS_FACTORY = {"doc_with_image": DocWithImageTool, "doc": DocTool}



# SPDX-FileCopyrightText: © 2023 Haystack <https://github.com/deepset-ai/haystack.git>
# SPDX-License-Identifier: Apache License 2.0

import collections
from typing import Any, Dict, List, Optional, OrderedDict


class ConversationMemory:
    """
    A memory class that stores conversation history.
    """

    def __init__(self, input_key: str = "input", output_key: str = "output", generator: Any = None):
        """
        Initialize ConversationMemory with input and output keys.

        :param input_key: The key to use for storing user input.
        :param output_key: The key to use for storing model output.
        """
        self.list: List[OrderedDict] = []
        self.input_key = input_key
        self.output_key = output_key
        self.generator = generator
        self.generator_modifies_user_text = hasattr(self.generator, "get_user_text")

    def get_user_text(self, chat_snippet):
        if self.generator_modifies_user_text:
            user_text = self.generator.get_user_text(chat_snippet)
        else:
            user_text = chat_snippet["Human"]
        return {"role": "user", "content": user_text}

    def load(self, keys: Optional[List[str]] = None, **kwargs) -> str:
        """
        Load conversation history as a formatted string.

        :param keys: Optional list of keys (ignored in this implementation).
        :param kwargs: Optional keyword arguments
            - window_size: integer specifying the number of most recent conversation snippets to load.
        :return: A formatted string containing the conversation history.
        """
        window_size = kwargs.get("window_size", None)

        if window_size is not None:
            chat_list = self.list[-window_size:]  # pylint: disable=invalid-unary-operand-type
        else:
            chat_list = self.list

        roles = []
        for chat_snippet in chat_list:
            user_role = self.get_user_text(chat_snippet)

            assistant_text = f"{chat_snippet['AI']}\n"
            # assistant_text = f"Assistant: {chat_snippet['AI']}\n"
            if "observation" in chat_snippet:
                assistant_text += f"Observation: {chat_snippet['observation']}\n"
            assistant_role = {"role": "assistant", "content": assistant_text}

            roles.extend([user_role, assistant_role])

        return roles

    def save(self, data: Dict[str, Any]) -> None:
        """
        Save a conversation snippet to memory.

        :param data: A dictionary containing the conversation snippet to save.
        """
        chat_snippet = collections.OrderedDict()
        chat_snippet["Human"] = data[self.input_key]
        chat_snippet["AI"] = data[self.output_key]

        for param_key in ["observation", "additional_params"]:
            if data.get(param_key, None):
                chat_snippet[param_key] = data[param_key]
        self.list.append(chat_snippet)

    def get_additional_params(self):
        params = {}
        for chat_snippet in self.list:
            additional_params = chat_snippet.get("additional_params", {})
            if len(additional_params) > 0:
                for key in params:
                    if key in additional_params:
                        params[key] += additional_params[key]

                for key in additional_params:
                    if key not in params:
                        params[key] = additional_params[key]
        return params

    def clear(self) -> None:
        """
        Clear the conversation history.
        """
        self.list = []



import logging
import os
import time
from argparse import Namespace
from typing import Any, Dict, List, Optional

import torch
from haystack import component
from haystack.components.generators import HuggingFaceLocalGenerator
from haystack.lazy_imports import LazyImport
from haystack.utils import ComponentDevice, Secret

with LazyImport("Install optimum-habana using 'pip install -e .[habana]'") as gaudi_import:
    from fastrag.generators.gaudi_utils import adjust_batch, initialize_model

logger = logging.getLogger()


def setup_parser(huggingface_pipeline_kwargs: dict, model_name_or_path: str):
    """
    Optimum-habana text generation paramters.
    """
    huggingface_pipeline_kwargs.setdefault(
        "static_shapes", True
    )  # whether to use static shapes, if possible
    huggingface_pipeline_kwargs.setdefault(
        "input_padding", False
    )  # whether to apply padding to the input_ids when max_input_tokens == 0

    #   For a detailed description of each parameter, please visit:
    #   https://github.com/huggingface/optimum-habana/blob/main/examples/text-generation/run_generation.py
    #   All settings are explained in the setup_parser function.

    huggingface_pipeline_kwargs.setdefault("device", "hpu")
    huggingface_pipeline_kwargs.setdefault("model_name_or_path", model_name_or_path)
    huggingface_pipeline_kwargs.setdefault("bf16", True)
    huggingface_pipeline_kwargs.setdefault("max_new_tokens", 100)
    huggingface_pipeline_kwargs.setdefault("max_input_tokens", 0)
    huggingface_pipeline_kwargs.setdefault("batch_size", 1)
    huggingface_pipeline_kwargs.setdefault("warmup", 3)
    huggingface_pipeline_kwargs.setdefault("n_iterations", 5)
    huggingface_pipeline_kwargs.setdefault("local_rank", 0)
    huggingface_pipeline_kwargs.setdefault("use_kv_cache", True)
    huggingface_pipeline_kwargs.setdefault("use_hpu_graphs", True)
    huggingface_pipeline_kwargs.setdefault("dataset_name", None)
    huggingface_pipeline_kwargs.setdefault("column_name", None)
    huggingface_pipeline_kwargs.setdefault("do_sample", True)
    huggingface_pipeline_kwargs.setdefault("num_beams", 1)
    huggingface_pipeline_kwargs.setdefault("trim_logits", True)
    huggingface_pipeline_kwargs.setdefault("seed", 27)
    huggingface_pipeline_kwargs.setdefault("profiling_warmup_steps", 0)
    huggingface_pipeline_kwargs.setdefault("profiling_steps", 0)
    huggingface_pipeline_kwargs.setdefault("prompt", None)
    huggingface_pipeline_kwargs.setdefault("bad_words", None)
    huggingface_pipeline_kwargs.setdefault("force_words", None)
    huggingface_pipeline_kwargs.setdefault("peft_model", None)
    huggingface_pipeline_kwargs.setdefault("num_return_sequences", 1)
    huggingface_pipeline_kwargs.setdefault("token", None)
    huggingface_pipeline_kwargs.setdefault("model_revision", "main")
    huggingface_pipeline_kwargs.setdefault("attn_softmax_bf16", True)
    huggingface_pipeline_kwargs.setdefault("output_dir", None)
    huggingface_pipeline_kwargs.setdefault("bucket_size", -1)
    huggingface_pipeline_kwargs.setdefault(
        "bucket_internal", False
    )  # set to False to work for the default case bucket_size = -1
    huggingface_pipeline_kwargs.setdefault("dataset_max_samples", -1)
    huggingface_pipeline_kwargs.setdefault("limit_hpu_graphs", True)
    huggingface_pipeline_kwargs.setdefault("reuse_cache", True)
    huggingface_pipeline_kwargs.setdefault("verbose_workers", True)
    huggingface_pipeline_kwargs.setdefault("simulate_dyn_prompt", None)
    huggingface_pipeline_kwargs.setdefault("reduce_recompile", False)
    huggingface_pipeline_kwargs.setdefault("fp8", True)
    huggingface_pipeline_kwargs.setdefault("use_flash_attention", True)
    huggingface_pipeline_kwargs.setdefault("flash_attention_recompute", True)
    huggingface_pipeline_kwargs.setdefault("flash_attention_causal_mask", True)
    huggingface_pipeline_kwargs.setdefault("book_source", True)
    huggingface_pipeline_kwargs.setdefault("torch_compile", True)
    huggingface_pipeline_kwargs.setdefault("ignore_eos", True)
    huggingface_pipeline_kwargs.setdefault("temperature", 1.0)
    huggingface_pipeline_kwargs.setdefault("top_p", 1.0)
    huggingface_pipeline_kwargs.setdefault("const_serialization_path", None)
    huggingface_pipeline_kwargs.setdefault("disk_offload", True)
    huggingface_pipeline_kwargs.setdefault("trust_remote_code", True)

    args = Namespace()
    for k, v in huggingface_pipeline_kwargs.items():
        setattr(args, k, v)

    if args.torch_compile:
        args.use_hpu_graphs = False

    if not args.use_hpu_graphs:
        args.limit_hpu_graphs = False

    args.quant_config = os.getenv("QUANT_CONFIG", "")
    if args.quant_config == "" and args.disk_offload:
        logger.info(
            "`--disk_offload` was tested only with fp8, it may not work with full precision. If error raises try to remove the --disk_offload flag."
        )
    return args


class GaudiGenerator(HuggingFaceLocalGenerator):
    """
    A generator class that uses IPEX (Intel PyTorch Extension) for faster inference on Xeon CPUs.

    Args:
        model_name_or_path (str): The name or path of the pre-trained model. Defaults to "meta-llama/Llama-2-7b-hf".
        device (Optional[ComponentDevice]): The device to run the generator on. Defaults to None.
        token (Optional[Secret]): The secret token for authentication. Defaults to Secret.from_env_var("HF_API_TOKEN", strict=False).
        generation_kwargs (Optional[Dict[str, Any]]): Additional generation arguments. Defaults to None.
        huggingface_pipeline_kwargs (Optional[Dict[str, Any]]): Additional arguments for the Hugging Face pipeline. Defaults to None.
        stop_words (Optional[List[str]]): A list of stop words to remove from the generated replies. Defaults to None.
        dtype (Optional[str]): The data type to use for inference. Defaults to "bfloat16".
        **kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        model_name_or_path: str = "microsoft/Phi-3-mini-128k-instruct",
        device: Optional[ComponentDevice] = None,
        token: Optional[Secret] = Secret.from_env_var("HF_API_TOKEN", strict=False),
        generation_kwargs: Optional[Dict[str, Any]] = None,
        huggingface_pipeline_kwargs: Optional[Dict[str, Any]] = None,
        stop_words: Optional[List[str]] = None,
        dtype: Optional[str] = "bfloat16",  ## bfloat16 of float32
        **kwargs,
    ):
        gaudi_import.check()
        super().__init__(
            model=model_name_or_path,
            device=device,
            token=token,
            generation_kwargs=generation_kwargs,
            huggingface_pipeline_kwargs=huggingface_pipeline_kwargs,
            stop_words=stop_words,
        )
        self.model_name_or_path = model_name_or_path
        self.model = None
        self.token = token
        self.stop_words = stop_words
        self.generation_kwargs = generation_kwargs or {}

    def warm_up(self):
        """
        Initializes the generator by loading the pre-trained model and tokenizer.
        """
        if self.model is None:
            self.args = setup_parser(self.huggingface_pipeline_kwargs, self.model_name_or_path)
            self.model, self.tokenizer, self.generation_config = initialize_model(self.args, logger)
            self.tokenizer.padding_side = "left"
            self.use_lazy_mode = True

            if self.args.torch_compile and self.model.config.model_type == "llama":
                self.use_lazy_mode = False

            import habana_frameworks.torch.hpu as torch_hpu

    def generate(self, input_sentences):
        """Generates sequences from the input sentences and returns them."""
        size = None

        encode_t0 = time.perf_counter()
        # Tokenization
        if self.args.max_input_tokens > 0:
            input_tokens = self.tokenizer.batch_encode_plus(
                input_sentences,
                return_tensors="pt",
                padding="max_length",
                max_length=self.args.max_input_tokens,
                truncation=True,
            )
        else:
            input_tokens = self.tokenizer.batch_encode_plus(
                input_sentences, return_tensors="pt", padding=self.args.input_padding
            )

        encode_duration = time.perf_counter() - encode_t0

        if size is not None:
            input_tokens = adjust_batch(input_tokens, size)

        for t in input_tokens:
            if torch.is_tensor(input_tokens[t]):
                input_tokens[t] = input_tokens[t].to(self.args.device)
        iteration_times = []
        generation_kwargs = {
            k: v for k, v in self.generation_kwargs.items() if k not in ["return_full_text"]
        }
        outputs = self.model.generate(
            **input_tokens,
            generation_config=self.generation_config,
            lazy_mode=self.use_lazy_mode,
            hpu_graphs=self.args.use_hpu_graphs,
            profiling_steps=self.args.profiling_steps,
            profiling_warmup_steps=self.args.profiling_warmup_steps,
            ignore_eos=self.args.ignore_eos,
            iteration_times=iteration_times,
            **generation_kwargs,
        ).cpu()
        first_token_time = iteration_times[0] + encode_duration
        logger.info(f"Time to first token = {first_token_time*1000}ms")
        outputs = outputs[:, input_tokens["input_ids"].shape[1] :]
        return self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

    @component.output_types(replies=List[str])
    def run(
        self,
        prompt: str,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, List[str]]:
        """
        Generates replies based on the given prompt.

        Args:
            prompt (str): The prompt for generating replies.
            generation_kwargs (Optional[Dict[str, Any]]): Additional generation arguments. Defaults to None.

        Returns:
            Dict[str, List[str]]: A dictionary containing the generated replies.
        """
        replies = self.generate([prompt])
        return {"replies": replies}



import ast
import hashlib
import math
from abc import abstractmethod
from typing import Dict, List

import pandas as pd
from datasets import load_dataset
from haystack import Document
from tqdm import tqdm

from fastrag.utils import AnswerGroundType, remove_html_from_text

try:
    import nltk

    nltk.download("punkt")
    nltk_tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
except ImportError as e:
    print("nltk is not installed")


class BaseParser:
    def __init__(self, batch_size, limit=None):
        # since document_store.write_documents *requires* batch_size, this parameter is required
        self.batch_size = batch_size
        self.limit = limit

    @abstractmethod
    def __iter__(self):
        pass


def text_encoder(row) -> Document:
    """encoder for text passages"""
    return Document(content=str(row["text"]))


def wikidpedia_data_encoder(row) -> Document:
    """encoder for wikipedia passages from HF"""
    row_id = str(hashlib.sha256(row["_id"].encode()).hexdigest())
    return Document(
        content=str(row["passage_text"]),
        id=row_id,
        meta={"title": str(row["article_title"])},
    )


def wiki_odqa_tasks_encoder(doc) -> Dict:
    """encoder for NQ/TQA ODQA dataset"""
    query = str(doc["query"])
    answers = list(set(doc["answers"]))
    return {"question": query, "answers": answers}


def squad_odqa_encoder(doc) -> Dict:
    """encoder for squad ODQA dataset"""
    query = str(doc["query"])
    answers = list(set(ast.literal_eval("".join(doc["answers"]))))
    return {"question": query, "answers": answers}


def wikipedia_hf_encoder(doc) -> Document:
    """encoder for wikipedia dataset from HF datasets"""
    return Document(content=doc["text"], id=doc["docid"], meta={"title": doc["title"]})


def pubmedQA_hf_encoder(doc) -> Document:
    """encoder for pubmedQA dataset from HF datasets

    Splits the original passage into subpassages 5 sentences each.
    """
    sentences = nltk_tokenizer.tokenize(doc["context"])
    joined_passages = sentences_to_passages(sentences, sentences_per_passage=5)

    return [
        Document(content=s, id=f"{str(doc['document_id'])}_{s_idx}")
        for s_idx, s in enumerate(joined_passages)
    ]


def hf_id_title_text(doc) -> Document:
    """encoder for BeIR/NQ dataset from HF datasets"""
    return Document(content=doc["text"], id=doc["_id"], meta={"title": doc["title"]})


def hf_id_title_text_concat(doc) -> Document:
    """encoder for BeIR/NQ dataset from HF datasets"""
    return Document(content=(doc["title"] + "\n" + doc["text"]), id=doc["_id"])


def sentences_to_passages(sens, sentences_per_passage=3):
    sens_count = len(sens)

    sens_batch_count = (sens_count // sentences_per_passage) + 1

    passages = []
    for sen_batch_index in range(sens_batch_count):
        sens_current_batch = sens[
            sen_batch_index * sentences_per_passage : (sen_batch_index + 1) * sentences_per_passage
        ]
        if len(sens_current_batch) == 0:
            continue

        passage_current = " ".join(sens_current_batch)
        passages.append(passage_current)

    return passages


def wikipedia_hf_multisentence_encoder(doc) -> List[Document]:
    """encoder for wikipedia dataset from HF datasets"""

    sentences = nltk_tokenizer.tokenize(str(doc["text"]))

    joined_passages = sentences_to_passages(sentences)

    return [
        Document(content=s, id=f"{str(doc['id'])}_{s_idx}", meta={"title": str(doc["title"])})
        for s_idx, s in enumerate(joined_passages)
    ]


encoding_methods = {
    "text": text_encoder,
    "wikidpedia_data": wikidpedia_data_encoder,
    "squad_odqa": squad_odqa_encoder,
    "nq_decoder": wiki_odqa_tasks_encoder,
    "tqa_decoder": wiki_odqa_tasks_encoder,
    "wikipedia_hf": wikipedia_hf_encoder,
    "wikipedia_hf_multisentence": wikipedia_hf_multisentence_encoder,
    "pubmedqa_hf": pubmedQA_hf_encoder,
    "hf_id_title_text": hf_id_title_text,
    "hf_id_title_text_concat": hf_id_title_text_concat,
}


class HFDatasetLoader(BaseParser):
    def __init__(
        self,
        dataset_info,
        encoding_method,
        batch_size=1,
        limit=None,
    ):
        super().__init__(batch_size, limit)

        # load the dataset from HF
        self.data = load_dataset(**dataset_info)
        self.length = self.limit or len(self.data)
        self.chunks = (self.length // self.batch_size) + 1
        self.encode_fn = encoding_methods[encoding_method]

    def __iter__(self):
        for i in tqdm(
            list(range(self.chunks)),
            desc="Data chunks",
        ):
            docs = self.process(i)
            yield docs

    def process(self, i):
        """Given a batch number i, returns the processed batch.
        Useful for restarting a job from the ith batch."""

        end_size = min((i + 1) * self.batch_size, self.length)
        docs = []
        for j in range(i * self.batch_size, end_size):
            encoding_results = self.encode_fn(self.data[j])
            if isinstance(encoding_results, list):
                docs.extend(encoding_results)
            else:
                docs.append(encoding_results)
        return docs


def encode_stackoverflow(row, fields_to_construct_from, id_field):
    """
    Encodes the content in the stackoverflow datasets, to create documents for the store.
    fields_to_construct_from: The field to use for creating the text.
    id_field: The field that corresponds with the id of the content.
    """
    contents = []
    for field in fields_to_construct_from:
        content = str(row[field])
        if field != "title":
            content = remove_html_from_text(content)
        contents.append(content)
    overall_text = ". ".join(contents)
    return Document(content=overall_text, id=str(id_field))


def encode_wikipedia_text(row):
    return Document(content=str(row["text"]), id=str(row["id"]))


def encode_wikipedia_text_and_title(row):
    return Document(
        content=str(row["text"]),
        id=str(row["id"]),
        meta={"title": str(row["title"])},
    )


def encode_wikipedia_title_only(row):
    return Document(content=str(row["title"]), id=str(row["id"]))


def encode_wikipedia_json(row):
    return Document(content=row["text"], id=str(row["id"]), meta={"title": row["title"]})


def encode_stackoverflow_answer(row):
    return encode_stackoverflow(
        row, fields_to_construct_from=["answer_body"], id_field="AcceptedAnswerId"
    )


def encode_stackoverflow_body_answer(row):
    return encode_stackoverflow(
        row, fields_to_construct_from=["body", "answer_body"], id_field="id"
    )


def encode_stackoverflow_body(row):
    return encode_stackoverflow(row, fields_to_construct_from=["body"], id_field="id")


def encode_question_answer_test_csv(row):
    return {"question": str(row[0]), "answers": list(set(ast.literal_eval(row[1])))}


def encode_evaluation_question_content(row):
    return {
        "question": str(row["title"]),
        "answers": [str(row["id"])],
        "answer_type": AnswerGroundType.ID,
    }


row_parser_functions = {
    "text": encode_wikipedia_text,
    "text_and_title": encode_wikipedia_text_and_title,
    "only_title": encode_wikipedia_title_only,
    "question_answer_test_csv": encode_question_answer_test_csv,
    "stackoverflow_answer": encode_stackoverflow_answer,
    "stackoverflow_body_answer": encode_stackoverflow_body_answer,
    "stackoverflow_body": encode_stackoverflow_body,
    "evaluation_question_content": encode_evaluation_question_content,
    "wikipedia_local_json": encode_wikipedia_json,
}


class CSVFileLoader(BaseParser):
    def __init__(
        self,
        filepath,
        encoding_method,
        delimiter="\t",
        batch_size=1,
        limit=None,
    ):
        super().__init__(batch_size, limit)

        self.filepath = filepath
        self.file_length = self.limit or _get_file_length(self.filepath)
        self.iterations_count = math.ceil(self.file_length / self.batch_size)
        self.delimiter = delimiter
        self.encode_fn = row_parser_functions[encoding_method]

    def __iter__(self):
        for df in tqdm(
            pd.read_csv(
                filepath_or_buffer=self.filepath,
                chunksize=self.batch_size,
                delimiter=self.delimiter,
            ),
            total=self.iterations_count,
            desc="Document chunks",
        ):
            docs = [self.encode_fn(row) for _, row in df.iterrows()]
            yield docs


class JSONFileLoader(BaseParser):
    def __init__(
        self,
        filepath,
        encoding_method,
        batch_size=1,
    ):
        super().__init__(batch_size)

        self.filepath = filepath
        self.file_length = _get_file_length(self.filepath)
        self.iterations_count = math.ceil(self.file_length / self.batch_size)
        self.encode_fn = row_parser_functions[encoding_method]

    def __iter__(self):
        for df in tqdm(
            pd.read_json(
                self.filepath,
                chunksize=self.batch_size,
                lines=self.filepath.endswith("jsonl"),
            ),
            total=self.iterations_count,
            desc="Document chunks",
        ):
            docs = [self.encode_fn(row) for _, row in df.iterrows()]
            yield docs


def _get_file_length(filepath):
    with open(filepath) as f:
        line_count = 0
        for line in f:
            line_count += 1
    return line_count



from pathlib import Path
from typing import List, Optional, Union

import torch
from haystack import ComponentError, Document, component
from haystack.components.embedders import SentenceTransformersDocumentEmbedder
from haystack.utils import ComponentDevice, Secret


@component
class BiEncoderSimilarityRanker:
    """
    A class that represents a Bi-Encoder Similarity Ranker.

    This ranker is used to rank a list of documents based on their similarity to a given query.

    Args:
        model (Union[str, Path]): The path to the model or the model name.
        device (Optional[ComponentDevice]): The device to run the model on. Defaults to None.
        token (Optional[Secret]): The authentication token. Defaults to Secret.from_env_var("HF_API_TOKEN", strict=False).
        top_k (int): The number of top results to retrieve. Defaults to 10.
        query_prefix (str): The prefix to add to the query input. Defaults to "".
        document_prefix (str): The prefix to add to the document input. Defaults to "".
        batch_size (int): The batch size for inference. Defaults to 32.
        normalize (bool): Whether to normalize the scores. Defaults to True.
        pooling (str): The pooling strategy for encoding. Defaults to "mean".
        meta_fields_to_embed (Optional[List[str]]): The list of meta fields to embed. Defaults to None.
        embedding_separator (str): The separator for embedding meta fields. Defaults to "\n".
    """

    def __init__(
        self,
        model: Union[str, Path],
        device: Optional[ComponentDevice] = None,
        token: Optional[Secret] = Secret.from_env_var("HF_API_TOKEN", strict=False),
        top_k: int = 10,
        query_prefix: str = "",
        document_prefix: str = "",
        batch_size: int = 32,
        normalize: bool = True,
        pooling: str = "mean",
        meta_fields_to_embed: Optional[List[str]] = None,
        embedding_separator: str = "\n",
    ):
        self.embedder_model = None

        self.model_name_or_path = model
        self.device = ComponentDevice.resolve_device(device)
        self.token = token
        self.top_k = top_k
        self.query_prefix = query_prefix
        self.document_prefix = document_prefix
        self.batch_size = batch_size
        self.normalize = normalize
        self.pooling = pooling
        self.meta_fields_to_embed = meta_fields_to_embed or []
        self.embedding_separator = embedding_separator

        if self.top_k <= 0:
            raise ValueError(f"top_k must be > 0, but got {top_k}")

    def warm_up(self):
        self.embedder_model = SentenceTransformersDocumentEmbedder(
            model=self.model_name_or_path,
            device=self.device,
            token=self.token,
            prefix=self.document_prefix,  ## we'll use only for document embeddings, query is one time
            batch_size=self.batch_size,
            normalize_embeddings=self.normalize,
            meta_fields_to_embed=self.meta_fields_to_embed,
            embedding_separator=self.embedding_separator,
        )
        self.embedder_model.warm_up()

    @component.output_types(documents=List[Document])
    def run(
        self,
        query: str,
        documents: List[Document],
        top_k: Optional[int] = None,
        score_threshold: Optional[float] = None,
    ):
        """
        Runs the ranking algorithm to rank the given documents based on the query.

        Args:
            query (str): The query string.
            documents (List[Document]): The list of documents to be ranked.
            top_k (Optional[int]): The maximum number of documents to return. If not provided, the default value will be used.
        """
        if not documents:
            return {"documents": []}

        top_k = top_k or self.top_k

        if top_k <= 0:
            raise ValueError(f"top_k must be > 0, but got {top_k}")

        if self.embedder_model is None:
            raise ComponentError(
                f"The component {self.__class__.__name__} wasn't warmed up. Run 'warm_up()' before calling 'run()'."
            )

        query_document = Document(content=self.query_prefix + query)
        query_vector = self.embedder_model.run([query_document])["documents"][0].embedding
        documents_with_vectors = self.embedder_model.run(documents=documents)["documents"]

        # Calculate similarity scores between query_document and documents_with_vectors
        doc_vectors = torch.tensor([doc.embedding for doc in documents_with_vectors])
        scores = torch.tensor(query_vector) @ doc_vectors.T  ## perhaps need to break it into chunks
        scores = scores.reshape(len(documents))
        indices = scores.cpu().sort(descending=True).indices
        return {"documents": [documents_with_vectors[i.item()] for i in indices[:top_k]]}



import argparse
import base64
import json
import os

import chainlit as cl
from chainlit.sync import run_sync
from events import Events
from haystack.components.generators.hugging_face_local import HuggingFaceLocalGenerator

from fastrag.agents.base import AgentTokenStreamingHandler, HFTokenStreamingHandler
from fastrag.agents.create_agent import get_basic_conversation_pipeline

config = os.environ.get("CONFIG", "config/rag_pipeline_chat.yaml")

args = argparse.Namespace(app_type="conversation", config=config)

generator, tools_objects = get_basic_conversation_pipeline(args)
callback_manager = Events()

stream_handler = AgentTokenStreamingHandler(callback_manager)
generator.generation_kwargs["streamer"] = HFTokenStreamingHandler(
    generator.pipeline.tokenizer, stream_handler
)

memory = [
    {
        "role": "system",
        "content": """You are a helpful assistant.
Your answers must be short and to the point.
""",
    }
]


@cl.on_chat_end
def chat_end():
    global memory
    # clear memory
    memory = []


@cl.on_message
async def main(message: cl.Message):
    global current_settings

    def parse_element(element, params):
        # insert the image into the params
        if "image" in element.mime:
            img_str = base64.b64encode(element.content).decode("utf-8")
            if "images" not in params:
                params["images"] = []
            params["images"].append(img_str)

        # insert the text into the params
        if "text" in element.mime:
            file_text = element.content.decode("utf-8")
            if "file_texts" not in params:
                params["file_texts"] = []
            params["file_texts"].append(file_text.split("\n"))

        if "json" in element.mime:
            if "data_rows" not in params:
                params["data_rows"] = []

            data_rows = json.load(open(element.path, "r"))
            params["data_rows"].append(data_rows)

    # params for the agent
    params = {}

    if len(message.elements) > 0:
        # parse the input elements, such as images, text, etc.
        for el in message.elements:
            parse_element(el, params)

        # Upload text and images, when appropriate
        for tool in tools_objects:
            if hasattr(tool, "upload_data_to_pipeline"):
                _ = await cl.Message(
                    author="Agent", content=f"Uploading into {tool.name}..."
                ).send()
                tool.upload_data_to_pipeline(params)

    user_query = message.content
    user_input = user_query
    for tool in tools_objects:
        tool_result = tool.run(user_query)
        user_input += tool_result
        _ = await cl.Message(author=tool.name, content=tool_result).send()

    memory.append({"role": "user", "content": user_input})

    prompt = generator.pipeline.tokenizer.apply_chat_template(
        memory, tokenize=False, add_generation_prompt=True
    )

    message = cl.Message(author="Agent", content="")

    def stream_to_message(token):
        run_sync(message.stream_token(token))

    callback_manager.on_new_token = stream_to_message

    result = await cl.make_async(generator.run)(prompt)

    answer = result["replies"][0]

    memory.append({"role": "assistant", "content": answer})



import argparse
import base64
import os
from io import BytesIO

import chainlit as cl
from chainlit.input_widget import Select, Slider

from fastrag.rest_api.conversation_pipeline_creator import get_conversation_pipeline

# Path to pipeline configuration (yaml).
# The YAML config file is formatted as follows:
#    chat_model:
#        model_kwargs: {}
#        model_name_or_path: liuhaotian/llava-v1-0719-336px-lora-merge-vicuna-13b-v1.3
#        use_gpu: true
#    summary_model:
#        model_kwargs: {}
#        model_name_or_path: meta-llama/Llama-2-7b-chat-hf
#        use_gpu: true
#    doc_pipeline_file: PATH_TO_DOC_PIPELINE_DILE
#    image_pipeline_file: PATH_TO_IMAGE_PIPELINE_DILE
config = os.environ.get("CONFIG", "config/visual_chat.yaml")
delete_on_end_chat = bool(int(os.environ.get("DELETE_ON_END_CHAT", "0")))

args = argparse.Namespace(app_type="conversation", config=config)

pipeline = get_conversation_pipeline(args)

cl.HaystackAgentCallbackHandler(pipeline.agent)


MAP_ACTIONS = {
    "Regular Message": dict(use_retrieval=False, use_image_retrieval=False),
    "Do Image Retrieval": dict(use_retrieval=False, use_image_retrieval=True),
    "Do Text Retrieval": dict(use_retrieval=True, use_image_retrieval=False),
    "Do Image + Text Retrieval": dict(use_retrieval=True, use_image_retrieval=True),
}

current_settings = None


@cl.on_chat_start
async def start():
    global current_settings
    settings = await cl.ChatSettings(
        [
            Select(
                id="message_action",
                label="Message Action",
                values=list(MAP_ACTIONS.keys()),
                initial_index=0,
            ),
            Slider(
                id="top_k_retriever",
                label="Documents to Retrieve",
                initial=10,
                min=1,
                max=1000,
                step=1,
            ),
            Slider(
                id="top_k_reranker",
                label="Documents to Re-Rank",
                initial=1,
                min=1,
                max=1000,
                step=1,
            ),
        ]
    ).send()

    current_settings = settings


@cl.on_settings_update
async def setup_agent(settings):
    global current_settings
    current_settings = settings


@cl.on_chat_end
def chat_end():
    pipeline.agent.memory.clear()
    if delete_on_end_chat:
        pipeline.delete_all_data({})


@cl.on_message
async def main(message: cl.Message):
    global current_settings

    def parse_element(element, params):
        # insert the image into the params
        if "image" in element.mime:
            img_str = base64.b64encode(element.content).decode("utf-8")
            if "images" not in params:
                params["images"] = []
            params["images"].append(img_str)

        # insert the text into the params
        if "text" in element.mime:
            file_text = element.content.decode("utf-8")
            if "file_texts" not in params:
                params["file_texts"] = []
            params["file_texts"].append(file_text.split("\n"))

    # params for the agent
    params = {
        "inputs": [],
        "ret_pipe": {
            "Retriever": {"top_k": current_settings["top_k_retriever"]},
            "Reranker": {"top_k": current_settings["top_k_reranker"]},
        },
        "debug": True,
    }

    if len(message.elements) > 0:
        # parse the input elements, such as images, text, etc.
        for el in message.elements:
            parse_element(el, params)

        # Upload text and images, when appropriate
        if "file_texts" in params:
            _ = await cl.Message(author="Agent", content="Uploading Text...").send()
            pipeline.upload_docs(params)

        if "images" in params:
            _ = await cl.Message(author="Agent", content="Uploading Images...").send()
            pipeline.upload_images(params)

    agent_result, returned_docs, _, images_to_return = await cl.make_async(pipeline.run)(
        message.content, params, **MAP_ACTIONS[current_settings["message_action"]]
    )
    answer = agent_result["answers"][0].answer

    # Send a message stating that an image has been retrieved
    if len(images_to_return) > 0:
        image_base64 = images_to_return["images"][0]

        bytes_io = BytesIO(base64.b64decode(image_base64))
        with open("imageToSave.png", "wb") as fh:
            fh.write(bytes_io.read())

        image_elements = [cl.Image(name="image1", display="inline", path="imageToSave.png")]

        _ = await cl.Message(
            content="Here is the image I have found.", elements=image_elements
        ).send()

    # Send a message stating that a text document has been retrieved
    if len(returned_docs):
        text_elements = [
            cl.Text(name=f"Document {doc_i+1}", content=doc.content, display="inline")
            for doc_i, doc in enumerate(returned_docs)
        ]

        await cl.Message(
            content="Here are the documents I have found.", elements=text_elements
        ).send()

    # Send the agent answer
    _ = await cl.Message(author="Agent", content=answer).send()



# coding=utf-8
# Copyright 2022 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
# Copyright (C) 2020-2021 Habana Labs, Ltd. an Intel Company
###############################################################################

import copy
import glob
import logging
import os
import shutil
import tempfile
import time
from pathlib import Path
from typing import List, Optional, Union

import torch
from optimum.habana.checkpoint_utils import (
    get_ds_injection_policy,
    get_repo_root,
    model_is_optimized,
    model_on_meta,
    write_checkpoints_json,
)
from optimum.habana.utils import (
    check_habana_frameworks_version,
    check_optimum_habana_min_version,
    set_seed,
)
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    EosTokenCriteria,
    GenerationConfig,
    MaxLengthCriteria,
    MaxTimeCriteria,
    StoppingCriteriaList,
    StopStringCriteria,
)
from transformers.utils import check_min_version


class NoRepeatEosTokenCriteria(EosTokenCriteria):
    """
    This class can be used to stop generation whenever the "end-of-sequence" token is generated.
    By default, it uses the `model.generation_config.eos_token_id`.

    Args:
        eos_token_id (`Union[int, List[int], torch.Tensor]`):
            The id(s) of the *end-of-sequence* token.
    """

    def __init__(self, eos_token_id: Union[int, List[int], torch.Tensor]):
        super().__init__(eos_token_id)
        self.previous_length = {}

    def is_last_eos_token(self, input_ids):
        tok = self.eos_token_id
        # tok_appear = (input_ids == tok).nonzero()
        no_eos_input_ids = (input_ids != tok).nonzero()

        for batch_index in range(input_ids.shape[0]):
            # get current batch
            batch_tok_appear = no_eos_input_ids[no_eos_input_ids[:, 0] == batch_index]

            # check if an eos has been added
            only_eos_added = False
            if hasattr(self, "previous_length"):
                if batch_index in self.previous_length:
                    # if an eos has been added, then the current length and the previous will be the same
                    # since after removing all the eos tokens, the number of regular tokens did not change.
                    only_eos_added = batch_tok_appear.shape[0] == self.previous_length[batch_index]
                self.previous_length[batch_index] = batch_tok_appear.shape[0]
            else:
                self.previous_length = {batch_index: batch_tok_appear.shape[0]}
            if only_eos_added:
                return True
        return False

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs
    ) -> torch.BoolTensor:
        if input_ids.device.type == "mps":
            return super().__call__(input_ids, scores, **kwargs)
        else:
            self.eos_token_id = self.eos_token_id.to(input_ids.device)
            return self.is_last_eos_token(input_ids)


def _get_stopping_criteria(
    self,
    generation_config: GenerationConfig,
    stopping_criteria: Optional[StoppingCriteriaList],
    tokenizer: Optional["PreTrainedTokenizerBase"] = None,
    **kwargs,
) -> StoppingCriteriaList:
    criteria = StoppingCriteriaList()
    if generation_config.max_length is not None:
        max_position_embeddings = getattr(self.config, "max_position_embeddings", None)
        criteria.append(
            MaxLengthCriteria(
                max_length=generation_config.max_length,
                max_position_embeddings=max_position_embeddings,
            )
        )
    if generation_config.max_time is not None:
        criteria.append(MaxTimeCriteria(max_time=generation_config.max_time))
    if generation_config.stop_strings is not None:
        if tokenizer is None:
            raise ValueError(
                "There are one or more stop strings, either in the arguments to `generate` or in the "
                "model's generation config, but we could not locate a tokenizer. When generating with "
                "stop strings, you must pass the model's tokenizer to the `tokenizer` argument of `generate`."
            )
        criteria.append(
            StopStringCriteria(stop_strings=generation_config.stop_strings, tokenizer=tokenizer)
        )
    if generation_config.eos_token_id is not None:
        criteria.append(NoRepeatEosTokenCriteria(eos_token_id=generation_config.eos_token_id))
    criteria = self._merge_criteria_processor_list(criteria, stopping_criteria)
    return criteria


def _get_generation_mode(self, generation_config, assistant_model=None):
    return GenerationMode.GREEDY_SEARCH


from transformers.generation.utils import GenerationMixin, GenerationMode

# replace EosTokenCriteria with NoRepeatEosTokenCriteria, ensuring that the eos tokens added to the
# end of the sequence do not stop the generation process

GenerationMixin._get_stopping_criteria = _get_stopping_criteria
GenerationMixin._get_generation_mode = _get_generation_mode


def adjust_batch(batch, size):
    curr_size = batch["input_ids"].shape[1]
    if curr_size >= size:
        adjusted_batch = {
            "input_ids": batch["input_ids"][:, :size],
            "attention_mask": batch["attention_mask"][:, :size],
        }
    else:
        adjusted_batch = {}
        for k in batch.keys():
            last_colm = batch[k][:, -1]
            expanded = last_colm.tile((size - curr_size, 1)).T
            adjusted_batch[k] = torch.concat([batch[k], expanded], 1)
    assert adjusted_batch["input_ids"].shape[1] == size
    assert adjusted_batch["attention_mask"].shape[1] == size
    return adjusted_batch


def override_print(enable):
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop("force", False)
        if force or enable:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def override_logger(logger, enable):
    logger_info = logger.info

    def info(*args, **kwargs):
        force = kwargs.pop("force", False)
        if force or enable:
            logger_info(*args, **kwargs)

    logger.info = info


def count_hpu_graphs():
    return len(glob.glob(".graph_dumps/*PreGraph*"))


def override_prints(enable, logger):
    override_print(enable)
    override_logger(logger, enable)


def setup_distributed(args):
    args.local_rank = int(os.getenv("LOCAL_RANK", "0"))
    args.world_size = int(os.getenv("WORLD_SIZE", "0"))
    args.global_rank = int(os.getenv("RANK", "0"))


def setup_const_serialization(const_serialization_path):
    import uuid

    const_serialization_path = os.path.join(const_serialization_path + uuid.uuid4().hex)
    os.makedirs(const_serialization_path)
    from habana_frameworks.torch.hpu import enable_const_section_serialization

    print("Serializing const params to {}".format(const_serialization_path))
    enable_const_section_serialization(const_serialization_path, True)


def setup_env(args):
    # Will error if the minimal version of Transformers is not installed. Remove at your own risks.
    check_min_version("4.34.0")
    check_optimum_habana_min_version("1.9.0.dev0")
    # TODO: SW-167588 - WA for memory issue in hqt prep_model
    os.environ.setdefault("EXPERIMENTAL_WEIGHT_SHARING", "FALSE")

    if args.global_rank == 0 and not args.torch_compile:
        os.environ.setdefault("GRAPH_VISUALIZATION", "true")
        shutil.rmtree(".graph_dumps", ignore_errors=True)

    if args.world_size > 0:
        os.environ.setdefault("PT_HPU_LAZY_ACC_PAR_MODE", "0")
        os.environ.setdefault("PT_HPU_ENABLE_LAZY_COLLECTIVES", "true")

    # Tweak generation so that it runs faster on Gaudi
    from optimum.habana.transformers.modeling_utils import adapt_transformers_to_gaudi

    adapt_transformers_to_gaudi()


def setup_device(args):
    if args.device == "hpu":
        import habana_frameworks.torch.core as htcore

        if args.fp8:
            htcore.hpu_set_env()
    return torch.device(args.device)


# patching LinearAllreduce to use ScopedLinearAllReduce
def patch_scoped_linear_all_reduce(model):
    from deepspeed.module_inject.layers import LinearAllreduce
    from optimum.habana.transformers.models.modeling_all_models import ScopedLinearAllReduce

    for name, module in model.named_children():
        if type(module) is LinearAllreduce:
            SL = ScopedLinearAllReduce(mod=module)
            setattr(model, name, SL)
        patch_scoped_linear_all_reduce(module)


def get_torch_compiled_model(model):
    model.model = torch.compile(model.model, backend="hpu_backend")
    return model


def setup_model(args, model_dtype, model_kwargs, logger):
    logger.info("Single-device run.")

    if args.disk_offload:
        from accelerate import infer_auto_device_map, init_empty_weights

        config = AutoConfig.from_pretrained(args.model_name_or_path)
        with init_empty_weights():
            model = AutoModelForCausalLM.from_config(config)
        max_memory = {"cpu": "10GiB"}
        device_map = infer_auto_device_map(model, max_memory=max_memory, dtype=model_dtype)
        model = AutoModelForCausalLM.from_pretrained(
            args.model_name_or_path,
            device_map=device_map,
            offload_folder="/tmp/offload_folder/",
            offload_state_dict=True,
            torch_dtype=model_dtype,
            **model_kwargs,
        )
    else:
        if args.peft_model is not None:
            model = peft_model(args, model_dtype, logger, **model_kwargs)
        else:
            model = AutoModelForCausalLM.from_pretrained(
                args.model_name_or_path, torch_dtype=model_dtype, **model_kwargs
            )
    if args.quant_config:
        import habana_quantization_toolkit

        habana_quantization_toolkit.prep_model(model)
    model = model.eval().to(args.device)

    if args.use_hpu_graphs:
        from habana_frameworks.torch.hpu import wrap_in_hpu_graph
        from optimum.habana.transformers.trainer import _is_peft_model

        if check_habana_frameworks_version("1.13.0") and model.config.model_type == "falcon":
            model = wrap_in_hpu_graph(model, hash_with_views=False)
        else:
            model = wrap_in_hpu_graph(model)
        if _is_peft_model(model):
            model.base_model = wrap_in_hpu_graph(model.base_model)

    if args.torch_compile and model.config.model_type == "llama":
        model = get_torch_compiled_model(model)

    return model


def setup_distributed_model(args, model_dtype, model_kwargs, logger):
    import deepspeed

    logger.info("DeepSpeed is enabled.")
    deepspeed.init_distributed(dist_backend="hccl")
    config = AutoConfig.from_pretrained(
        args.model_name_or_path, torch_dtype=model_dtype, **model_kwargs
    )
    load_to_meta = model_on_meta(config)

    if load_to_meta:
        # Construct model with fake meta tensors, later will be replaced on devices during ds-inference ckpt load
        with deepspeed.OnDevice(dtype=model_dtype, device="meta"):
            model = AutoModelForCausalLM.from_config(config, torch_dtype=model_dtype)

        # Model loaded to meta is managed differently
        checkpoints_json = tempfile.NamedTemporaryFile(suffix=".json", mode="+w")

        # For PEFT models, write the merged model on disk to be able to load it on the meta device
        if args.peft_model is not None:
            merged_model_dir = "/tmp/text_generation_merged_peft_model"
            if args.local_rank == 0:
                if Path(merged_model_dir).is_dir():
                    shutil.rmtree(merged_model_dir)
                peft_model(args, model_dtype, logger, **model_kwargs).save_pretrained(
                    merged_model_dir
                )
            torch.distributed.barrier()

        write_checkpoints_json(
            merged_model_dir if args.peft_model is not None else args.model_name_or_path,
            args.local_rank,
            checkpoints_json,
            token=args.token,
        )
    else:
        # TODO: revisit placement on CPU when auto-injection is possible
        with deepspeed.OnDevice(dtype=model_dtype, device="cpu"):
            if args.peft_model is not None:
                model = peft_model(args, model_dtype, logger, **model_kwargs)
            else:
                model = AutoModelForCausalLM.from_pretrained(
                    args.model_name_or_path, torch_dtype=model_dtype, **model_kwargs
                )
    model.eval()

    # Initialize the model
    ds_inference_kwargs = {"dtype": model_dtype}
    ds_inference_kwargs["tensor_parallel"] = {"tp_size": args.world_size}
    ds_inference_kwargs["enable_cuda_graph"] = args.use_hpu_graphs
    ds_inference_kwargs["injection_policy"] = get_ds_injection_policy(config)
    if load_to_meta:
        ds_inference_kwargs["checkpoint"] = checkpoints_json.name

    model = deepspeed.init_inference(model, **ds_inference_kwargs)
    model = model.module
    if model.config.model_type in ["llama", "falcon"]:
        patch_scoped_linear_all_reduce(model)

    if args.quant_config:
        import habana_quantization_toolkit

        habana_quantization_toolkit.prep_model(model)

    if args.torch_compile and model.config.model_type == "llama":
        model = get_torch_compiled_model(model)

    return model


def peft_model(args, model_dtype, logger, **model_kwargs):
    import importlib.util

    if importlib.util.find_spec("peft") is None:
        raise ImportError("The `peft` package is not installed, please run: `pip install peft`.")
    from peft import AutoPeftModelForCausalLM
    from peft.config import PeftConfigMixin

    base_model_name = PeftConfigMixin.from_pretrained(
        args.peft_model,
        token=model_kwargs["token"] if "token" in model_kwargs else None,
    ).base_model_name_or_path

    base_model_is_local = Path(base_model_name).is_dir()
    if not base_model_is_local:
        # Check if the base model path to a remote repository on the HF Hub exists
        from huggingface_hub import list_repo_files

        try:
            list_repo_files(base_model_name)
            base_model_is_remote = True
        except Exception:
            base_model_is_remote = False

    if base_model_is_local or base_model_is_remote:
        model = AutoPeftModelForCausalLM.from_pretrained(
            args.peft_model, torch_dtype=model_dtype, **model_kwargs
        )
    else:
        # Since the base model doesn't exist locally nor remotely, use `args.model_name_or_path` as the base model
        logger.warning(
            f"The base model `{base_model_name}` of the LoRA configuration associated"
            f" to `{args.peft_model}` does not exist locally or remotely. Using "
            f"`--model_name_or_path {args.model_name_or_path}` as a fall back for the base model."
        )
        from peft import PeftModel

        model = AutoModelForCausalLM.from_pretrained(
            args.model_name_or_path, torch_dtype=model_dtype, **model_kwargs
        )
        model = PeftModel.from_pretrained(
            model, args.peft_model, torch_dtype=model_dtype, **model_kwargs
        )
    if hasattr(model, "merge_and_unload"):
        model = model.merge_and_unload()
        if model_dtype == torch.bfloat16:
            model = model.to(torch.bfloat16)
        return model
    else:
        from optimum.habana.peft.peft_model import (
            gaudi_generate,
            gaudi_prepare_inputs_for_generation,
        )

        model.__class__.generate = gaudi_generate
        model.__class__.prepare_inputs_for_generation = gaudi_prepare_inputs_for_generation
        return model


def setup_tokenizer(args, model):
    tokenizer_kwargs = {
        "revision": args.model_revision,
        "token": args.token,
        "trust_remote_code": args.trust_remote_code,
    }
    if args.bad_words is not None or args.force_words is not None:
        tokenizer_kwargs["add_prefix_space"] = True
    tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path, **tokenizer_kwargs)
    if not model.config.is_encoder_decoder:
        tokenizer.padding_side = "left"

    if model.config.model_type == "llama":
        # unwind broken decapoda-research config
        model.generation_config.pad_token_id = 0
        model.generation_config.bos_token_id = 1
        model.generation_config.eos_token_id = 2
        tokenizer.bos_token_id = model.generation_config.bos_token_id
        tokenizer.eos_token_id = model.generation_config.eos_token_id
        tokenizer.pad_token_id = model.generation_config.pad_token_id
        tokenizer.pad_token = tokenizer.decode(tokenizer.pad_token_id)
        tokenizer.eos_token = tokenizer.decode(tokenizer.eos_token_id)
        tokenizer.bos_token = tokenizer.decode(tokenizer.bos_token_id)
    if model.config.model_type == "persimmon":
        model.generation_config.pad_token_id = model.generation_config.eos_token_id
        tokenizer.bos_token_id = model.generation_config.bos_token_id
        tokenizer.eos_token_id = model.generation_config.eos_token_id
        tokenizer.pad_token_id = model.generation_config.pad_token_id
        tokenizer.pad_token = tokenizer.decode(tokenizer.pad_token_id)
        tokenizer.eos_token = tokenizer.decode(tokenizer.eos_token_id)
        tokenizer.bos_token = tokenizer.decode(tokenizer.bos_token_id)

    # Some models like GPT2 do not have a PAD token so we have to set it if necessary
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token
        model.generation_config.pad_token_id = model.generation_config.eos_token_id

    return tokenizer, model


def setup_generation_config(args, model, tokenizer):
    bad_words_ids = None
    force_words_ids = None
    if args.bad_words is not None:
        bad_words_ids = [
            tokenizer.encode(bad_word, add_special_tokens=False) for bad_word in args.bad_words
        ]
    if args.force_words is not None:
        force_words_ids = [
            tokenizer.encode(force_word, add_special_tokens=False)
            for force_word in args.force_words
        ]

    is_optimized = model_is_optimized(model.config)
    # Generation configuration
    generation_config = copy.deepcopy(model.generation_config)
    generation_config.max_new_tokens = args.max_new_tokens
    generation_config.use_cache = args.use_kv_cache
    generation_config.static_shapes = is_optimized and args.static_shapes
    generation_config.bucket_size = args.bucket_size if is_optimized else -1
    generation_config.bucket_internal = args.bucket_internal
    generation_config.do_sample = args.do_sample
    generation_config.num_beams = args.num_beams
    generation_config.bad_words_ids = bad_words_ids
    generation_config.force_words_ids = force_words_ids
    generation_config.num_return_sequences = args.num_return_sequences
    generation_config.trim_logits = args.trim_logits
    generation_config.attn_softmax_bf16 = args.attn_softmax_bf16
    generation_config.limit_hpu_graphs = args.limit_hpu_graphs
    generation_config.reuse_cache = args.reuse_cache
    generation_config.reduce_recompile = args.reduce_recompile
    if generation_config.reduce_recompile:
        assert generation_config.bucket_size > 0
    generation_config.use_flash_attention = args.use_flash_attention
    generation_config.flash_attention_recompute = args.flash_attention_recompute
    generation_config.flash_attention_causal_mask = args.flash_attention_causal_mask
    generation_config.trust_remote_code = args.trust_remote_code
    return generation_config


def initialize_model(args, logger):
    init_start = time.perf_counter()
    setup_distributed(args)
    override_prints(args.global_rank == 0 or args.verbose_workers, logger)
    setup_env(args)
    setup_device(args)
    set_seed(args.seed)
    get_repo_root(args.model_name_or_path, local_rank=args.local_rank, token=args.token)
    use_deepspeed = args.world_size > 0
    if use_deepspeed or args.bf16 or args.fp8:
        model_dtype = torch.bfloat16
    else:
        model_dtype = torch.float
        args.attn_softmax_bf16 = False

    model_kwargs = {
        "revision": args.model_revision,
        "token": args.token,
        "trust_remote_code": args.trust_remote_code,
    }
    if args.trust_remote_code:
        logger.warning(
            "`trust_remote_code` is set, there is no guarantee this model works properly and it may fail"
        )

    model = (
        setup_model(args, model_dtype, model_kwargs, logger)
        if not use_deepspeed
        else setup_distributed_model(args, model_dtype, model_kwargs, logger)
    )
    tokenizer, model = setup_tokenizer(args, model)
    generation_config = setup_generation_config(args, model, tokenizer)

    if args.const_serialization_path:
        setup_const_serialization(args.const_serialization_path)
    if args.fp8:
        import habana_frameworks.torch.core as htcore

        print("Initializing inference mode")
        const_marking = os.getenv("ENABLE_CONST_MARKING", "True")
        if const_marking == "True":
            htcore.hpu_initialize(model)
    init_end = time.perf_counter()
    logger.info(f"Args: {args}")
    logger.info(
        f"device: {args.device}, n_hpu: {args.world_size}, bf16: {model_dtype == torch.bfloat16}"
    )
    logger.info(f"Model initialization took {(init_end - init_start):.3f}s")
    return model, tokenizer, generation_config



from typing import Dict, Tuple

from haystack import Pipeline
from haystack.components.writers import DocumentWriter

from fastrag.agents.utils import load_text

######## Query Handlers

COMPONENT_WITH_STORE = "retriever"


class QueryHandler:
    def query(self, query, params) -> Tuple[str, Dict]:
        observation: str = ""
        additional_params: dict = {}
        return observation, additional_params


class HaystackQueryHandler(QueryHandler):
    def __init__(self, pipeline, query_function=None, component_with_store=COMPONENT_WITH_STORE):
        self.pipeline = pipeline
        self.query_function = query_function if query_function else self.default_query_function
        self.component_with_store = component_with_store

    def default_query_function(self, query, params):
        """
        Create the default querying of the pipeline, by inserting the input query into all mandatory fields.
        """
        pipe_inputs = self.pipeline.inputs()
        query_dict = {}
        for inp_node_name, inp_node_params in pipe_inputs.items():
            for param_name, param_values in inp_node_params.items():
                if param_values["is_mandatory"]:
                    if inp_node_name not in query_dict:
                        query_dict[inp_node_name] = {}

                    query_dict[inp_node_name][param_name] = query

        return query_dict

    def query(self, query, params) -> Tuple[str, Dict]:
        run_params = self.query_function(query, params)
        return self.pipeline.run(run_params)

    def get_store(self):
        return self.pipeline.get_component(self.component_with_store).document_store


class HaystackFileQueryHandler(HaystackQueryHandler):
    def __init__(
        self, pipeline_yaml_path, query_function=None, component_with_store=COMPONENT_WITH_STORE
    ):
        pipeline = Pipeline.loads(load_text(pipeline_yaml_path))
        super().__init__(pipeline, query_function, component_with_store)


######## Indexing Handlers


class IndexingHandler:
    def index(self, docs):
        pass


class HaystackIndexingHandler(IndexingHandler):
    def __init__(self, document_store, indexing_function=None, create_pipeline=None):
        """
        :param create_pipeline:
            A function that creates the indexing pipeline. We create it to avoid problems with stale pipelines
            in haystack v2.
        :param indexing_function:
            A function that indexes a given list of documents into the indexing pipeline.
        """
        if create_pipeline:
            self.create_pipeline = create_pipeline
        self.indexing_function = (
            indexing_function if indexing_function else self.default_indexing_function
        )
        self.document_store = document_store

    def index(self, docs):
        pipeline = self.create_pipeline()
        self.indexing_function(pipeline, self.document_store, docs)

    def default_indexing_function(self, indexing_pipeline, document_store, documents):
        indexing_pipeline.add_component(
            instance=DocumentWriter(document_store=document_store), name="doc_writer"
        )

        indexing_pipeline.connect("doc_embedder.documents", "doc_writer.documents")

        indexing_pipeline.run({"doc_embedder": {"documents": documents}})


class HaystackFileIndexingHandler(HaystackIndexingHandler):
    def __init__(self, pipeline_yaml_path, document_store, indexing_function=None):
        super().__init__(document_store, indexing_function)
        self.pipeline_yaml = load_text(pipeline_yaml_path)

    def create_pipeline(self):
        return Pipeline.loads(self.pipeline_yaml)


QUERY_HANDLER_FACTORY = {
    "haystack": HaystackQueryHandler,
    "haystack_yaml": HaystackFileQueryHandler,
}

INDEX_HANDLER_FACTORY = {
    "haystack": HaystackIndexingHandler,
    "haystack_yaml": HaystackFileIndexingHandler,
}



import logging
import warnings
from typing import Any, Dict, List, Optional, Union

import torch
import torch.distributed as dist
import transformers
from haystack import Document, component
from haystack.components.generators import HuggingFaceLocalGenerator
from haystack.utils import ComponentDevice
from torch import nn
from transformers import (
    MODEL_FOR_CAUSAL_LM_MAPPING,
    AutoConfig,
    AutoTokenizer,
    GenerationMixin,
    LogitsProcessor,
    LogitsProcessorList,
    StoppingCriteriaList,
)
from transformers.generation.stopping_criteria import validate_stopping_criteria
from transformers.generation.utils import (
    SampleDecoderOnlyOutput,
    SampleEncoderDecoderOutput,
    SampleOutput,
)

logger = logging.getLogger(__name__)


class REPLUG_Generation(GenerationMixin):
    """Implementing REPLUG-based sampling text generation."""

    def sample(
        self,
        input_ids: torch.LongTensor,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        logits_warper: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[Union[int, List[int]]] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: bool = False,
        streamer: Optional["BaseStreamer"] = None,
        **model_kwargs,
    ) -> Union[SampleOutput, torch.LongTensor]:
        # init values
        logits_processor = (
            logits_processor if logits_processor is not None else LogitsProcessorList()
        )
        stopping_criteria = (
            stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
        )
        if max_length is not None:
            warnings.warn(
                "`max_length` is deprecated in this function, use"
                " `stopping_criteria=StoppingCriteriaList(MaxLengthCriteria(max_length=max_length))` instead.",
                UserWarning,
            )
            stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
        logits_warper = logits_warper if logits_warper is not None else LogitsProcessorList()
        pad_token_id = (
            pad_token_id if pad_token_id is not None else self.generation_config.pad_token_id
        )
        eos_token_id = (
            eos_token_id if eos_token_id is not None else self.generation_config.eos_token_id
        )
        if isinstance(eos_token_id, int):
            eos_token_id = [eos_token_id]
        eos_token_id_tensor = (
            torch.tensor(eos_token_id).to(input_ids.device) if eos_token_id is not None else None
        )
        output_scores = (
            output_scores if output_scores is not None else self.generation_config.output_scores
        )
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.generation_config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.generation_config.output_hidden_states
        )
        return_dict_in_generate = (
            return_dict_in_generate
            if return_dict_in_generate is not None
            else self.generation_config.return_dict_in_generate
        )

        # init attention / hidden states / scores tuples
        scores = () if (return_dict_in_generate and output_scores) else None
        decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
        cross_attentions = () if (return_dict_in_generate and output_attentions) else None
        decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

        # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
        if return_dict_in_generate and self.config.is_encoder_decoder:
            encoder_attentions = (
                model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
            )
            encoder_hidden_states = (
                model_kwargs["encoder_outputs"].get("hidden_states")
                if output_hidden_states
                else None
            )

        # keep track of which sequences are already finished
        unfinished_sequences = torch.ones(
            input_ids.shape[0], dtype=torch.long, device=input_ids.device
        )

        this_peer_finished = False  # used by synced_gpus only
        # auto-regressive generation
        while True:
            if synced_gpus:
                # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
                # The following logic allows an early break if all peers finished generating their sequence
                this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(
                    input_ids.device
                )
                # send 0.0 if we finished, 1.0 otherwise
                dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
                # did all peers finish? the reduced sum will be 0.0 then
                if this_peer_finished_flag.item() == 0.0:
                    break

            # prepare model inputs
            model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)

            # forward pass to get next token
            outputs = self(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
            )

            if synced_gpus and this_peer_finished:
                continue  # don't waste resources running the code we don't need

            next_token_logits = outputs.logits[:, -1, :]

            # pre-process distribution
            ### REPLUG - document weighting is done via REPLUGLogitsProcessor
            next_token_scores = logits_processor(input_ids, next_token_logits)
            next_token_scores = logits_warper(input_ids, next_token_scores)

            # Store scores, attentions and hidden_states when required
            if return_dict_in_generate:
                if output_scores:
                    scores += (next_token_scores,)
                if output_attentions:
                    decoder_attentions += (
                        (outputs.decoder_attentions,)
                        if self.config.is_encoder_decoder
                        else (outputs.attentions,)
                    )
                    if self.config.is_encoder_decoder:
                        cross_attentions += (outputs.cross_attentions,)

                if output_hidden_states:
                    decoder_hidden_states += (
                        (outputs.decoder_hidden_states,)
                        if self.config.is_encoder_decoder
                        else (outputs.hidden_states,)
                    )

            ### REPLUG
            # Sample from the normalized "logits", assuming the REPLUG processor was used!
            probs = nn.functional.softmax(next_token_scores, dim=-1)
            next_tokens = torch.multinomial(probs, num_samples=1).squeeze(-1)
            # Lock same next-token for all examples in batch
            next_tokens[:] = next_tokens[0]

            # finished sentences should have their next token be a padding token
            if eos_token_id is not None:
                if pad_token_id is None:
                    raise ValueError(
                        "If `eos_token_id` is defined, make sure that `pad_token_id` is defined."
                    )
                next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
                    1 - unfinished_sequences
                )

            # update generated ids, model inputs, and length for next step
            input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
            if streamer is not None:
                streamer.put(next_tokens.cpu())
            model_kwargs = self._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
            )

            # if eos_token was found in one sentence, set sentence to finished
            if eos_token_id_tensor is not None:
                unfinished_sequences = unfinished_sequences.mul(
                    next_tokens.tile(eos_token_id_tensor.shape[0], 1)
                    .ne(eos_token_id_tensor.unsqueeze(1))
                    .prod(dim=0)
                )

                # stop when each sentence is finished
                if unfinished_sequences.max() == 0:
                    this_peer_finished = True

            # stop if we exceed the maximum length
            if stopping_criteria(input_ids, scores):
                this_peer_finished = True

            if this_peer_finished and not synced_gpus:
                break

        if streamer is not None:
            streamer.end()

        if return_dict_in_generate:
            if self.config.is_encoder_decoder:
                return SampleEncoderDecoderOutput(
                    sequences=input_ids,
                    scores=scores,
                    encoder_attentions=encoder_attentions,
                    encoder_hidden_states=encoder_hidden_states,
                    decoder_attentions=decoder_attentions,
                    cross_attentions=cross_attentions,
                    decoder_hidden_states=decoder_hidden_states,
                )
            else:
                return SampleDecoderOnlyOutput(
                    sequences=input_ids,
                    scores=scores,
                    attentions=decoder_attentions,
                    hidden_states=decoder_hidden_states,
                )
        else:
            return input_ids


class REPLUGLogitsProcessor(LogitsProcessor):
    """
    REPLUG processor uses documents similarity scores to weight the different examples in a batch.
    """

    def __init__(self, documents_scores: torch.FloatTensor):
        "Documents' scores assumed to be L1 normalized."
        self.num_docs = documents_scores.shape[0]
        self.doc_scores = torch.unsqueeze(documents_scores, 1)

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        "Scores assumed to be logits. Returns logits averaged over the batch."
        replug_scores = self.doc_scores * scores
        replug_scores = replug_scores.sum(dim=0)
        replug_scores = torch.tile(replug_scores, (self.num_docs, 1))
        return replug_scores


class HF_REPLUG:
    "Creates a HF model that inherits from REPLUG_Generation class"

    def __new__(cls, name_or_path, **kwargs):
        return factory(name_or_path).from_pretrained(name_or_path, **kwargs)


def factory(name_or_path):
    loadedConfig = AutoConfig.from_pretrained(name_or_path)
    try:
        pretrained_class_object = getattr(transformers, loadedConfig.architectures[0])
        if pretrained_class_object not in MODEL_FOR_CAUSAL_LM_MAPPING.values():
            raise ValueError(
                f"Model {pretrained_class_object} is not used for causal LM generation."
            )
    except AttributeError:
        raise ValueError("Transformers architecture unknown.")

    class HF(pretrained_class_object, REPLUG_Generation):
        """Wrapper around HuggingFace transformers with REPLUG generation."""

        _keys_to_ignore_on_load_unexpected = [r"cls"]
        _tied_weights_keys = ["lm_head.weight"]

    return HF


@component
class ReplugGenerator(HuggingFaceLocalGenerator):
    """
    A subclass of the HuggingFaceLocalGenerator class. It loads a pre-trained model from Hugging Face, and uses
    utilizes the model according to the REPLUG algorithm (https://arxiv.org/abs/2301.12652)
    """

    def __init__(
        self,
        model: str = "meta-llama/Llama-2-7b-chat-hf",
        device: Optional[ComponentDevice] = None,
        generation_kwargs: Optional[Dict[str, Any]] = None,
        huggingface_pipeline_kwargs: Optional[Dict[str, Any]] = {},
        doc_sep_token: str = "###REPLUG-DOC###",
        **kwargs,
    ):
        super(ReplugGenerator, self).__init__(
            model=model,
            device=device,
            generation_kwargs=generation_kwargs,
            huggingface_pipeline_kwargs=huggingface_pipeline_kwargs.copy(),
            **kwargs,
        )

        self.model = model
        self.generation_kwargs.pop("return_full_text", None)
        self.orig_kwargs = huggingface_pipeline_kwargs

        tokenizer = AutoTokenizer.from_pretrained(model)
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.padding_side = "left"
        self.huggingface_pipeline_kwargs["tokenizer"] = tokenizer

        self.doc_sep_token = doc_sep_token

    def warm_up(self):
        """
        Initializes the component.
        """
        if self.pipeline is None:
            model_instance = HF_REPLUG(self.model, **self.orig_kwargs)
            self.huggingface_pipeline_kwargs["model"] = model_instance
            super(ReplugGenerator, self).warm_up()

    def create_doc_batch_prompt(self, prompt: str, documents: List[Document]):
        "Creating REPLUG prompts using a single prompt and a list of haystack documents."
        prompts = [prompt.replace(self.doc_sep_token, doc.content) for doc in documents]
        return prompts

    @component.output_types(replies=List[str])
    def run(
        self,
        prompt: str,
        documents: List[Document],
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Run the text generation model on the given prompt.

        :param prompt:
            A string representing the prompt.
        :param generation_kwargs:
            Additional keyword arguments for text generation.

        :returns:
            A dictionary containing the generated replies.
            - replies: A list of strings representing the generated replies.
        """
        if self.pipeline is None:
            raise RuntimeError(
                "The generation model has not been loaded. Please call warm_up() before running."
            )

        if not prompt:
            return {"replies": []}

        # merge generation kwargs from init method with those from run method
        updated_generation_kwargs = {**self.generation_kwargs, **(generation_kwargs or {})}

        ### REPLUG - retrieved documents processing
        prompts = self.create_doc_batch_prompt(prompt, documents)

        scores_norm = torch.tensor([doc.score for doc in documents], dtype=torch.float32).to(
            self.pipeline.model.device
        )
        scores_norm /= scores_norm.sum()  # L1 normalization

        inputs = self.pipeline.tokenizer(
            prompts, truncation=True, padding=True, return_tensors="pt"
        ).to(self.pipeline.model.device)

        gen_outputs = self.pipeline.model.generate(
            **inputs,
            stopping_criteria=self.stopping_criteria_list,
            logits_processor=LogitsProcessorList([REPLUGLogitsProcessor(scores_norm)]),
            **updated_generation_kwargs,
        )

        generated_tokens = gen_outputs[0][inputs["input_ids"].shape[1] :]
        replies = self.pipeline.tokenizer.decode(generated_tokens, skip_special_tokens=True)

        if self.stop_words:
            # the output of the pipeline includes the stop word
            replies = [
                reply.replace(stop_word, "").rstrip()
                for reply in replies
                for stop_word in self.stop_words
            ]

        return {"replies": replies}



from fastrag.rankers.bi_encoder_ranker import BiEncoderSimilarityRanker
from fastrag.rankers.colbert import ColBERTRanker
from fastrag.rankers.ipex_bi_encoder_ranker import IPEXBiEncoderSimilarityRanker

__all__ = [
    "BiEncoderSimilarityRanker",
    "IPEXBiEncoderSimilarityRanker",
]



import re
from enum import Enum
from functools import wraps
from importlib import import_module
from time import perf_counter
from typing import Any, Dict, Optional

import yaml


# adapted from https://github.com/deepset-ai/haystack/blob/594d2a10f84d13aef495c1cfbdaf4acad730c914/haystack/nodes/ranker/base.py#L68
def fastrag_timing(self, fn, attr_name):
    """Wrapper method used to time functions."""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if attr_name not in self.__dict__:
            self.__dict__[attr_name] = 0
        tic = perf_counter()
        ret = fn(*args, **kwargs)
        toc = perf_counter()
        self.__dict__[attr_name] = toc - tic
        return ret

    return wrapper


def add_timing_to_pipeline(fn):
    """wrapper for Haystack Pipeline to add timing info"""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        ret = fn(*args, **kwargs)
        p = args[0]
        # print(args, kwargs)
        ret["timings"] = get_timing_from_pipeline(p)
        return ret

    return wrapper


def get_timing_from_pipeline(pipeline):
    """run print_time() in each component that has timing measurement func"""
    timings = {}
    for c in pipeline.components:
        if hasattr(pipeline.components[c], "query_time"):
            name = pipeline.components[c].__class__.__name__
            timings[name] = (
                pipeline.components[c].query_count,
                pipeline.components[c].query_time,
            )
    return timings


def missing_deps(classname: str, import_error: Exception):
    class MissingDependency:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                f"Failed to import '{classname}', " f"(Original error: {str(import_error)})"
            ) from import_error

        def __getattr__(self, *a, **k):
            return None

    return MissingDependency


def init_haystack_cls(component_name: str, parameters: Dict[str, Any], name: Optional[str] = None):
    return BaseComponent._create_instance(component_name, parameters, name)


def init_cls(class_name: str, parameters: dict, **kwargs):
    """Initialize a class object given name as string and initiate using parameters dict"""
    try:
        module_path, class_name = class_name.rsplit(".", 1)
        module = import_module(module_path)
        class_init = getattr(module, class_name)

    except (ImportError, AttributeError) as e:
        raise ImportError(class_name)
    try:
        p = parameters.copy()
        p.update(kwargs)
        new_cls_obj = class_init(**p)
        return new_cls_obj
    except Exception as e:
        raise e


def load_yaml(filename: str) -> dict:
    """Load a yaml from file and return as dicts"""
    with open(filename) as fp:
        content = yaml.safe_load(fp)
    return content


def regex_match(text, pattern):
    """Test if a regex pattern is contained within a text."""
    try:
        pattern = re.compile(pattern, flags=re.IGNORECASE + re.UNICODE + re.MULTILINE)
    except BaseException:
        return False
    return pattern.search(text) is not None


GET_TEXT_MAP = {
    "direct": lambda x: x["text"],
    "content": lambda x: x.content,
}

GET_TITLE_MAP = {
    "direct": lambda x: x["title"],
    "content": lambda x: x.meta["title"],
}


def get_ans(text, answers):
    for ans in answers:
        has_answer = regex_match(text, ans)
        if has_answer:
            return True
    return False


def get_retrieval_gt_by_ids(contexts, expected):
    return [int(x.id in expected) for x in contexts]


def get_has_answer_data(contexts, answers, answer_type, text_get="direct", title_get="direct"):
    if answer_type == AnswerGroundType.ID:
        return get_retrieval_gt_by_ids(contexts, answers)

    passage_ground_truth = []
    for c in contexts:
        text = GET_TEXT_MAP[text_get](c)
        title = GET_TITLE_MAP[title_get](c)
        context_has_answer = get_ans(text, answers)
        title_has_answer = get_ans(title, answers)
        passage_has_answer = int(context_has_answer or title_has_answer)
        passage_ground_truth.append(passage_has_answer)

    return passage_ground_truth


def remove_html_from_text(text):
    text = re.sub("<[^<]+?>", "", text)
    text = text.replace("\n", " ")
    return text


class AnswerGroundType(Enum):
    HAS_ANSWER = 1
    ID = 2



from typing import List, Union

import torch
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast, StoppingCriteria


class StopWordsByTextCriteria(StoppingCriteria):
    """
    Stops text generation if any one of the stop words is generated.

    Note: When a stop word is encountered, the generation of new text is stopped.
    However, if the stop word is in the prompt itself, it can stop generating new text
    prematurely after the first token. This is particularly important for LLMs designed
    for dialogue generation. For these models, like for example mosaicml/mpt-7b-chat,
    the output includes both the new text and the original prompt. Therefore, it's important
    to make sure your prompt has no stop words.
    """

    def __init__(
        self,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        stop_words: List[str],
        device: Union[str, torch.device] = "cpu",
        token_backwards_buffer: int = 4,  # number of tokens to convert to a string, in addition to the stop words tokens
    ):
        super().__init__()
        self.stop_words_text = stop_words
        self.tokenizer = tokenizer
        encoded_stop_words = tokenizer(
            stop_words, add_special_tokens=False, padding=True, return_tensors="pt"
        )
        self.stop_words = encoded_stop_words.input_ids.to(device)
        self.token_backwards_buffer = token_backwards_buffer

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        for stop_word in self.stop_words:
            found_stop_word = self.is_stop_word_found(input_ids, stop_word)
            if found_stop_word:
                return True
        return False

    def is_stop_word_found(self, generated_text_ids: torch.Tensor, stop_word: torch.Tensor) -> bool:
        generated_text_ids = generated_text_ids[-1]
        len_generated_text_ids = generated_text_ids.size(0)
        len_stop_word = stop_word.size(0)
        len_stop_word_with_buffer = (
            len_stop_word + self.token_backwards_buffer
        )  # include a bit more text backwards, to ensure if the stop word indeed occurs or not

        tokens_to_decode = generated_text_ids
        if len_generated_text_ids > len_stop_word_with_buffer:
            tokens_to_decode = generated_text_ids[
                len_generated_text_ids - len_stop_word_with_buffer :
            ]

        gen_text = self.tokenizer.decode(tokens_to_decode)
        return any([gen_text.endswith(sw) for sw in self.stop_words_text])



import json
import os
from typing import Any

from haystack.lazy_imports import LazyImport

from fastrag.prompt_compressors.llm_lingua import LLMLinguaPromptCompressor

with LazyImport("Install openvino using 'pip install -e .[openvino]'") as ov_import:
    from optimum.intel import OVModelForTokenClassification


class OVLLMLinguaPromptCompressor(LLMLinguaPromptCompressor):
    """
    OVLLMLinguaPromptCompressor is a component that compresses the prompt it receives, while preserving the most essential
    information within it.
    For more information, we refer to the original repo: https://github.com/microsoft/LLMLingua

    Usage example:
    ```python
    compressor = OVLLMLinguaPromptCompressor()
    ```
    """

    def __init__(
        self,
        model_name_or_path: Any = "microsoft/llmlingua-2-xlm-roberta-large-meetingbank",
        device_map="cpu",
        rate=0.33,
        force_tokens=["\n", "?"],
        target_tokens=None,
    ):
        ov_import.check()
        model_config = json.load(open(os.path.join(model_name_or_path, "config.json")))
        super().__init__(
            model_name_or_path=model_config["_name_or_path"],
            device_map=device_map,
            rate=rate,
            force_tokens=force_tokens,
            target_tokens=target_tokens,
        )
        self.ov_model_name_or_path = model_name_or_path

    def warm_up(self):
        """
        Initializes the component.
        """
        super().warm_up()
        self.llm_lingua.model = OVModelForTokenClassification.from_pretrained(
            self.ov_model_name_or_path
        )



import logging
from typing import List, Optional

from haystack import Document, component

from fastrag.stores import PLAIDDocumentStore

logger = logging.getLogger(__name__)


@component
class ColBERTRetriever:
    def __init__(
        self,
        document_store: PLAIDDocumentStore,
        top_k: int = 10,
    ):
        self.document_store = document_store
        self.top_k = top_k

        logger.info(f"Initialize retriever using the store: {document_store}")

    @component.output_types(documents=List[Document])
    def run(self, query: str, top_k: Optional[int] = None, **kwargs) -> List[Document]:
        top_k = top_k or self.top_k
        documents = self.document_store.query(query_str=query, top_k=top_k)
        return {"documents": documents}



import argparse
import base64
import json
import os
import uuid
from io import BytesIO

import chainlit as cl
from chainlit_agent_fastrag_callback import HaystackAgentCallbackHandler

from fastrag.agents.create_agent import get_agent_conversation_pipeline

# Path to pipeline configuration (yaml).
# An example for a YAML config file is formatted as follows:
# chat_model:
#   generator_kwargs:
#       model: llava-hf/llava-1.5-7b-hf
#       task: "image-to-text"
#       generation_kwargs:
#         max_new_tokens: 100
#       huggingface_pipeline_kwargs:
#         torch_dtype: torch.bfloat16
#   generator_class: fastrag.generators.llava.LlavaHFGenerator
# tools:
#   - type: doc_with_image
#     path: "config/empty_retrieval_pipeline.yaml"
#     index_path: "config/empty_index_pipeline.yaml"
#     params:
#       name: "docRetriever"
#       description: "useful for when you need to retrieve textual documents to answer questions."
#       output_variable: "documents"
config = os.environ.get("CONFIG", "config/visual_chat_agent.yaml")
# delete_on_end_chat = bool(int(os.environ.get("DELETE_ON_END_CHAT", "0")))

args = argparse.Namespace(app_type="conversation", config=config)

agent = get_agent_conversation_pipeline(args)

HaystackAgentCallbackHandler(agent)


@cl.on_chat_end
def chat_end():
    # clear memory
    agent.memory.clear()


def add_images_to_message(additional_params):
    image_elements = []
    last_image = [additional_params["images"][-1]]
    for image_base64_index, image_base64 in enumerate(last_image):
        image_uuid = str(uuid.uuid4())
        bytes_io = BytesIO(base64.b64decode(image_base64))
        image_path = f"imageToSave_{image_base64_index}_{image_uuid}.png"
        with open(image_path, "wb") as fh:
            fh.write(bytes_io.read())

        image_elements.append(cl.Image(name=image_path, display="inline", path=image_path))
    return image_elements


def remove_images(image_elements):
    for e in image_elements:
        os.remove(e.path)


@cl.on_message
async def main(message: cl.Message):
    global current_settings

    def parse_element(element, params):
        # insert the image into the params
        if "image" in element.mime:
            img_str = base64.b64encode(element.content).decode("utf-8")
            if "images" not in params:
                params["images"] = []
            params["images"].append(img_str)

        # insert the text into the params
        if "text" in element.mime:
            file_text = element.content.decode("utf-8")
            if "file_texts" not in params:
                params["file_texts"] = []
            params["file_texts"].append(file_text.split("\n"))

        if "json" in element.mime:
            if "data_rows" not in params:
                params["data_rows"] = []

            data_rows = json.load(open(element.path, "r"))
            params["data_rows"].append(data_rows)

    # params for the agent
    params = {}

    if len(message.elements) > 0:
        # parse the input elements, such as images, text, etc.
        for el in message.elements:
            parse_element(el, params)

        # Upload text and images, when appropriate
        for tool_name in agent.tm.tools:
            tool = agent.tm.tools[tool_name]
            if hasattr(tool, "upload_data_to_pipeline"):
                _ = await cl.Message(
                    author="Agent", content=f"Uploading into {tool.name}..."
                ).send()
                tool.upload_data_to_pipeline(params)

    agent_result = await cl.make_async(agent.run)(message.content)
    answer = agent_result["answers"][0].answer

    # display retrieved image, if exists
    additional_params = agent.memory.get_additional_params()
    if "images" in additional_params and len(additional_params["images"]) > 0:
        image_elements = add_images_to_message(additional_params)

        _ = await cl.Message(
            content="Here are the images I have found.", elements=image_elements
        ).send()
        remove_images(image_elements)

    # Send the agent answer
    _ = await cl.Message(author="Agent", content=answer).send()



# SPDX-FileCopyrightText: © 2023 Haystack <https://github.com/deepset-ai/haystack.git>
# SPDX-License-Identifier: Apache License 2.0

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional

from fastrag.agents.agent_step import AgentStep

if TYPE_CHECKING:
    from fastrag.agents import Agent


class Color(Enum):
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\x1b[0m"


def load_text(path):
    return open(path, "r").read()


def clean_for_prompt(text):
    # ensure that {} do not appear in the prompt, to avoid format error.
    return text.replace("{", "{{").replace("}", "}}")


def print_text(text: str, end="", color: Optional[Color] = None) -> None:
    """
    Print text with optional color.
    :param text: Text to print.
    :param end: End character to use (defaults to "").
    :param color: Color to print text in (defaults to None).
    """
    if color:
        print(f"{color.value}{text}{Color.RESET.value}", end=end, flush=True)
    else:
        print(text, end=end, flush=True)


def react_parameter_resolver(
    query: str, agent: "Agent", agent_step: AgentStep, **kwargs
) -> Dict[str, Any]:
    """
    A parameter resolver for ReAct-based conversational agent that returns the query, the tool names, the tool names
    with descriptions, the history of the conversation, and the transcript (internal monologue).
    """
    return {
        "query": query,
        "tool_names_with_descriptions": agent.tm.get_tool_names_with_descriptions(),
        "transcript": agent_step.transcript,
        "memory": agent.memory.load(),
    }



from .fid import FiDGenerator
from .gaudi import GaudiGenerator
from .ipex import IPEXGenerator
from .llava import LlavaHFGenerator
from .openvino import OpenVINOGenerator
from .ort import ORTGenerator
from .replug import ReplugGenerator



import os

import chainlit as cl
import yaml
from chainlit.sync import run_sync
from events import Events
from transformers import AutoTokenizer, pipeline

from fastrag.agents.base import AgentTokenStreamingHandler, HFTokenStreamingHandler

config_path = os.environ.get("CONFIG", "config/regular_chat.yaml")
model_kwargs = yaml.safe_load(open(config_path, "r"))

callback_manager = Events()

model_name_or_path = model_kwargs["model"]

tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
stream_handler = AgentTokenStreamingHandler(callback_manager)
streamer = HFTokenStreamingHandler(tokenizer, stream_handler)
model_pipeline = pipeline(**model_kwargs, return_full_text=False, streamer=streamer)

memory = [
    {
        "role": "system",
        "content": """You are a helpful assistant.
Your answers must be short and to the point.
""",
    }
]


@cl.on_chat_end
def chat_end():
    global memory
    # clear memory
    memory = []


@cl.on_message
async def main(message: cl.Message):
    global current_settings

    memory.append({"role": "user", "content": message.content})

    prompt = model_pipeline.tokenizer.apply_chat_template(
        memory, tokenize=False, add_generation_prompt=True
    )

    message = cl.Message(author="Agent", content="")

    def stream_to_message(token):
        run_sync(message.stream_token(token))

    callback_manager.on_new_token = stream_to_message

    result = await cl.make_async(model_pipeline)(prompt)

    answer = result[0]["generated_text"]

    memory.append({"role": "assistant", "content": answer})



from typing import Any

from haystack import component
from haystack.lazy_imports import LazyImport

with LazyImport("Install llmlingua using 'pip install -e .[llmlingua]'") as llmlingua_import:
    from llmlingua import PromptCompressor


@component
class LLMLinguaPromptCompressor:
    """
    LLMLinguaPromptCompressor is a component that compresses the prompt it receives, while preserving the most essential
    information within it.
    For more information, we refer to the original repo: https://github.com/microsoft/LLMLingua

    Usage example:
    ```python
    compressor = LLMLinguaPromptCompressor()
    ```
    """

    def __init__(
        self,
        model_name_or_path: Any = "microsoft/llmlingua-2-xlm-roberta-large-meetingbank",
        device_map="cpu",
        rate=0.33,
        force_tokens=["\n", "?"],
        target_tokens=None,
    ):
        """
        Constructs a LLMLinguaPromptCompressor component.

        """
        llmlingua_import.check()
        self.rate = rate
        self.force_tokens = force_tokens
        self.target_tokens = target_tokens
        self.model_name_or_path = model_name_or_path
        self.device_map = device_map
        component.set_input_type(self, "prompt", Any, "")

    def warm_up(self):
        """
        Initializes the component.
        """
        self.llm_lingua = PromptCompressor(
            model_name=self.model_name_or_path,
            use_llmlingua2=True,
            device_map=self.device_map,
        )

    @component.output_types(prompt=str)
    def run(self, prompt: str, **kwargs):
        """
        :param kwargs:
            The variables that will be used to render the prompt template.

        :returns: A dictionary with the following keys:
            - `prompt`: The updated prompt text after rendering the prompt template.
        """
        compressed_prompt = self.llm_lingua.compress_prompt(
            prompt, rate=self.rate, force_tokens=self.force_tokens
        )

        return {"prompt": compressed_prompt["compressed_prompt"]}



from pathlib import Path
from typing import List, Optional, Union

import torch
from haystack import Document, component
from haystack.lazy_imports import LazyImport

with LazyImport("Run 'pip install .[colbert]' to install the ColBERT library.") as colbert_import:
    from colbert.infra import ColBERTConfig
    from colbert.modeling.checkpoint import Checkpoint
    from colbert.search.strided_tensor_core import StridedTensorCore


@component
class ColBERTRanker:
    def __init__(
        self,
        checkpoint_path: Union[str, Path],
        top_k: int = 10,
        query_maxlen: int = 120,
        doc_maxlen: int = 120,
        dim: int = 128,
        use_gpu: bool = False,
    ):
        colbert_import.check()
        self.top_k = top_k

        self.config = ColBERTConfig(
            checkpoint=checkpoint_path,
            ncells=2,
            nbits=2,
            query_maxlen=query_maxlen,
            doc_maxlen=doc_maxlen,
            dim=dim,
        )

        self.use_gpu = use_gpu
        self.to_cpu = not use_gpu

        self.checkpoint = Checkpoint(checkpoint_path, self.config)
        if self.use_gpu:
            self.checkpoint.model.cuda()

        self.checkpoint.query_tokenizer.query_maxlen = query_maxlen
        self.checkpoint.query_tokenizer.doc_maxlen = doc_maxlen

    @component.output_types(documents=List[Document])
    def run(
        self,
        query: str,
        documents: List[Document],
        top_k: Optional[int] = None,
    ) -> List[Document]:
        if top_k is None:
            top_k = self.top_k

        query_emb = self.checkpoint.queryFromText([query], bsize=64, to_cpu=self.to_cpu)
        if self.use_gpu:
            query_emb = query_emb.half()

        tensor, lengths = self.checkpoint.docFromText(
            [d.content for d in documents], keep_dims="flatten", bsize=64, to_cpu=self.to_cpu
        )
        tensor, masks = StridedTensorCore(tensor, lengths, use_gpu=self.use_gpu).as_padded_tensor()
        if not self.use_gpu:
            tensor, masks = tensor.float(), masks.float()

        # MaxSim implementation. Returns [#queries, #documents] tensor.
        # Output identical to colbert.modeling.colbert.colbert_score
        # however scores are different than when calculated at query time.
        # It's due to quantization differences when recovering the embeddings.
        scores = torch.einsum("bwd,qvd -> bqwv", query_emb, tensor * masks).max(-1).values.sum(-1)

        indices = scores.cpu().sort(descending=True).indices[0]
        return {"documents": [documents[i.item()] for i in indices[:top_k]]}
